# decorators.py
# Path: appointment/decorators.py

"""
Author: Adams Pierre David
Since: 2.0.0
"""

from functools import wraps

from appointment.utils.error_codes import ErrorCode
from appointment.utils.json_context import json_response
from appointment.utils.view_helpers import is_ajax


def require_user_authenticated(func):
    """Decorator to require a user to be authenticated.
    Usage: @require_user_authenticated
    """

    @wraps(func)
    def wrapper(request, *args, **kwargs):
        if not request.user.is_authenticated:
            return json_response("Not authorized.", status=401, success=False, error_code=ErrorCode.NOT_AUTHORIZED)
        return func(request, *args, **kwargs)

    return wrapper


def require_staff_or_superuser(func):
    """Decorator to require a user to be a staff or superuser.
    Usage: @require_staff_or_superuser
    """

    @wraps(func)
    def wrapper(request, *args, **kwargs):
        if not (request.user.is_staff or request.user.is_superuser):
            return json_response("Not authorized.", status=403, success=False, error_code=ErrorCode.NOT_AUTHORIZED)
        return func(request, *args, **kwargs)

    return wrapper


def require_superuser(func):
    """Decorator to require a user to be a superuser.
    Usage: @require_superuser
    """

    @wraps(func)
    def wrapper(request, *args, **kwargs):
        if not request.user.is_superuser:
            return json_response("Not authorized.", status=403, success=False, error_code=ErrorCode.NOT_AUTHORIZED)
        return func(request, *args, **kwargs)

    return wrapper


def require_ajax(func):
    """Decorator to require a request to be AJAX.
    Usage: @require_ajax
    """

    @wraps(func)
    def wrapper(request, *args, **kwargs):
        if not is_ajax(request):
            return json_response("Not an AJAX request.", status=400, success=False, error_code=ErrorCode.INVALID_DATA)
        return func(request, *args, **kwargs)

    return wrapper



# tasks.py
# Path: appointment/tasks.py

"""
Author: Adams Pierre David
Since: 3.1.0
"""
from django.utils.translation import gettext as _

from appointment.email_sender import notify_admin, send_email
from appointment.logger_config import logger
from appointment.models import Appointment


def send_email_reminder(to_email, first_name, reschedule_link, appointment_id):
    """
    Send a reminder email to the client about the upcoming appointment.
    """

    # Fetch the appointment using appointment_id
    logger.info(f"Sending reminder to {to_email} for appointment {appointment_id}")
    appointment = Appointment.objects.get(id=appointment_id)
    recipient_type = 'client'
    email_context = {
        'first_name': first_name,
        'appointment': appointment,
        'reschedule_link': reschedule_link,
        'recipient_type': recipient_type,
    }
    send_email(
        recipient_list=[to_email], subject=_("Reminder: Upcoming Appointment"),
        template_url='email_sender/reminder_email.html', context=email_context
    )
    # Notify the admin
    email_context['recipient_type'] = 'admin'
    notify_admin(
        subject=_("Admin Reminder: Upcoming Appointment"),
        template_url='email_sender/reminder_email.html', context=email_context
    )



# test_settings.py
# Path: appointment/tests/test_settings.py

from unittest.mock import patch

from django.test import TestCase

from appointment.settings import check_q_cluster


class CheckQClusterTest(TestCase):
    @patch('appointment.settings.settings')
    @patch('appointment.settings.logger')
    def test_check_q_cluster_with_django_q_missing(self, mock_logger, mock_settings):
        # Simulate 'django_q' not being in INSTALLED_APPS
        mock_settings.INSTALLED_APPS = []

        # Call the function under test
        result = check_q_cluster()

        # Check the result
        self.assertFalse(result)
        # Verify logger was called with the expected warning about 'django_q' not being installed
        mock_logger.warning.assert_called_with(
            "Django Q is not in settings.INSTALLED_APPS. Please add it to the list.\n"
            "Example: \n\n"
            "INSTALLED_APPS = [\n"
            "    ...\n"
            "    'appointment',\n"
            "    'django_q',\n"
            "]\n")

    @patch('appointment.settings.settings')
    @patch('appointment.settings.logger')
    def test_check_q_cluster_with_all_configurations_present(self, mock_logger, mock_settings):
        # Simulate both 'django_q' being in INSTALLED_APPS and 'Q_CLUSTER' configuration present
        mock_settings.INSTALLED_APPS = ['django_q']
        mock_settings.Q_CLUSTER = {'name': 'DjangORM'}

        # Call the function under test
        result = check_q_cluster()

        # Check the result and ensure no warnings are logged
        self.assertTrue(result)
        mock_logger.warning.assert_not_called()



import datetime
from datetime import timedelta

from django.db import IntegrityError
from django.test import TestCase
from django.utils.translation import gettext as _

from appointment.models import DayOff, Service, StaffMember, WorkingHours
from appointment.tests.mixins.base_mixin import ConfigMixin, ServiceMixin, StaffMemberMixin, UserMixin


class StaffMemberModelTestCase(TestCase, UserMixin, ServiceMixin, StaffMemberMixin, ConfigMixin):
    def setUp(self):
        self.user = self.create_user_()
        self.service = self.create_service_()
        self.staff_member = self.create_staff_member_(self.user, self.service)
        self.config = self.create_config_(lead_time=datetime.time(9, 0), finish_time=datetime.time(17, 0),
                                          slot_duration=30)

    def test_staff_member_creation(self):
        """Test if a staff member can be created."""
        self.assertIsNotNone(self.staff_member)
        self.assertEqual(self.staff_member.user, self.user)
        self.assertEqual(list(self.staff_member.get_services_offered()), [self.service])
        self.assertIsNone(self.staff_member.lead_time)
        self.assertIsNone(self.staff_member.finish_time)
        self.assertIsNone(self.staff_member.slot_duration)
        self.assertIsNone(self.staff_member.appointment_buffer_time)

    def test_staff_member_without_user(self):
        """A staff member cannot be created without a user."""
        with self.assertRaises(IntegrityError):
            StaffMember.objects.create()

    def test_staff_member_without_service(self):
        """A staff member can be created without a service."""
        self.staff_member.delete()
        new_staff_member = StaffMember.objects.create(user=self.user)
        self.assertIsNotNone(new_staff_member)
        self.assertEqual(new_staff_member.services_offered.count(), 0)

    def test_date_joined_auto_creation(self):
        """Test if the date_joined field is automatically set upon creation."""
        self.assertIsNotNone(self.staff_member.created_at)

    # Edge cases
    def test_staff_member_multiple_services(self):
        """A staff member can offer multiple services."""
        service2 = Service.objects.create(name="Test Service 2", duration=timedelta(hours=2), price=200)
        self.staff_member.services_offered.add(service2)
        self.assertIn(service2, self.staff_member.services_offered.all())

    def test_staff_member_with_non_existent_service(self):
        """A staff member cannot offer a non-existent service."""
        # Create a new staff member without any services
        self.staff_member.delete()
        new_staff_member = StaffMember.objects.create(user=self.user)

        # Trying to add a non-existent service to the staff member's services_offered
        with self.assertRaises(ValueError):
            new_staff_member.services_offered.add(
                Service(id=9999, name="Non-existent Service", duration=timedelta(hours=2), price=200))

    def test_str_representation(self):
        """Test the string representation of a StaffMember."""
        expected_str = self.staff_member.get_staff_member_name()
        self.assertEqual(str(self.staff_member), expected_str)

    def test_get_slot_duration_with_config(self):
        """Test get_slot_duration method with Config set."""
        self.config.slot_duration = 30
        self.config.save()
        self.assertEqual(self.staff_member.get_slot_duration(), 30)

    def test_get_slot_duration_text(self):
        """Test get_slot_duration_text method."""
        self.staff_member.slot_duration = 45
        self.assertEqual(self.staff_member.get_slot_duration_text(), "45 minutes")

    def test_get_lead_time(self):
        """Test get_lead_time method."""
        self.config.lead_time = datetime.time(9, 0)
        self.config.save()
        self.assertIsNone(self.staff_member.lead_time)
        self.assertEqual(self.staff_member.get_lead_time(), datetime.time(9, 0))

    def test_works_on_both_weekends_day(self):
        """Test works_on_both_weekends_day method."""
        self.staff_member.work_on_saturday = True
        self.staff_member.work_on_sunday = True
        self.assertTrue(self.staff_member.works_on_both_weekends_day())

    def test_get_non_working_days(self):
        """Test get_non_working_days method."""
        self.staff_member.work_on_saturday = False
        self.staff_member.work_on_sunday = False
        self.assertEqual(self.staff_member.get_non_working_days(),
                         [6, 0])  # [6, 0] represents Saturday and Sunday

    def test_get_services_offered(self):
        """Test get_services_offered method."""
        self.assertIn(self.service, self.staff_member.get_services_offered())

    def test_get_service_offered_text(self):
        """Test get_service_offered_text method."""
        self.assertEqual(self.staff_member.get_service_offered_text(), self.service.name)

    def test_get_appointment_buffer_time(self):
        """Test get_appointment_buffer_time method."""
        self.config.appointment_buffer_time = 15
        self.config.save()
        self.assertIsNone(self.staff_member.appointment_buffer_time)
        self.assertEqual(self.staff_member.get_appointment_buffer_time(), 15)

    def test_get_finish_time(self):
        """Test that the finish time is correctly returned from staff member or config."""
        # Case 1: Staff member has a defined finish time
        self.staff_member.finish_time = datetime.time(18, 0)
        self.staff_member.save()
        self.assertEqual(self.staff_member.get_finish_time(), datetime.time(18, 0))

        # Case 2: Staff member does not have a defined finish time, use config's
        self.staff_member.finish_time = None
        self.staff_member.save()
        self.assertEqual(self.staff_member.get_finish_time(), self.config.finish_time)

    def test_get_staff_member_first_name(self):
        """Test that the staff member's first name is returned."""
        self.assertEqual(self.staff_member.get_staff_member_first_name(), self.user.first_name)

    def test_get_weekend_days_worked_text(self):
        """Test various combinations of weekend work."""
        self.staff_member.work_on_saturday = True
        self.staff_member.work_on_sunday = False
        self.staff_member.save()
        self.assertEqual(self.staff_member.get_weekend_days_worked_text(), _("Saturday"))

        self.staff_member.work_on_sunday = True
        self.staff_member.save()
        self.assertEqual(self.staff_member.get_weekend_days_worked_text(), _("Saturday and Sunday"))

        self.staff_member.work_on_saturday = False
        self.staff_member.save()
        self.assertEqual(self.staff_member.get_weekend_days_worked_text(), _("Sunday"))

        self.staff_member.work_on_saturday = False
        self.staff_member.work_on_sunday = False
        self.staff_member.save()
        self.assertEqual(self.staff_member.get_weekend_days_worked_text(), _("None"))

    def test_get_appointment_buffer_time_text(self):
        """Test the textual representation of the appointment buffer time."""
        self.staff_member.appointment_buffer_time = 45  # 45 minutes
        self.assertEqual(self.staff_member.get_appointment_buffer_time_text(), "45 minutes")

    def test_get_days_off(self):
        """Test retrieval of days off."""
        DayOff.objects.create(staff_member=self.staff_member, start_date="2023-01-01", end_date="2023-01-02")
        self.assertEqual(len(self.staff_member.get_days_off()), 1)

    def test_get_working_hours(self):
        """Test retrieval of working hours."""
        WorkingHours.objects.create(staff_member=self.staff_member, day_of_week=1, start_time=datetime.time(9, 0),
                                    end_time=datetime.time(17, 0))
        self.assertEqual(len(self.staff_member.get_working_hours()), 1)

    def test_update_upon_working_hours_deletion(self):
        """Test the update of work_on_saturday and work_on_sunday upon working hours deletion."""
        self.staff_member.update_upon_working_hours_deletion(6)
        self.assertFalse(self.staff_member.work_on_saturday)
        self.staff_member.update_upon_working_hours_deletion(0)
        self.assertFalse(self.staff_member.work_on_sunday)

    def test_is_working_day(self):
        """Test whether a day is considered a working day."""
        self.staff_member.work_on_saturday = False
        self.staff_member.work_on_sunday = False
        self.staff_member.save()
        self.assertFalse(self.staff_member.is_working_day(6))  # Saturday
        self.assertTrue(self.staff_member.is_working_day(1))  # Monday



# test_utils.py
# Path: appointment/tests/test_utils.py

import datetime
import logging
from unittest.mock import patch

from django.apps import apps
from django.conf import settings
from django.http import HttpRequest
from django.test import TestCase
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from appointment.models import Appointment, AppointmentRequest, Config, Service, StaffMember
from appointment.services import get_available_slots, get_finish_button_text
from appointment.settings import APPOINTMENT_BUFFER_TIME, APPOINTMENT_FINISH_TIME, APPOINTMENT_LEAD_TIME, \
    APPOINTMENT_SLOT_DURATION, APPOINTMENT_WEBSITE_NAME
from appointment.utils.date_time import combine_date_and_time, convert_str_to_date, get_current_year, get_timestamp, \
    get_timezone
from appointment.utils.db_helpers import get_appointment_buffer_time, get_appointment_finish_time, \
    get_appointment_lead_time, get_appointment_slot_duration, get_user_model, get_website_name
from appointment.utils.view_helpers import generate_random_id, get_locale, is_ajax


class UtilityTestCase(TestCase):
    # Test cases for generate_random_id

    def setUp(self) -> None:
        self.test_service = Service.objects.create(name="Test Service", duration=datetime.timedelta(hours=1), price=100)
        self.user_model = get_user_model()
        self.user = self.user_model.objects.create_user(first_name="Tester",
                                                        email="testemail@gmail.com",
                                                        username="test_user", password="Kfdqi3!?n")
        self.staff_member = StaffMember.objects.create(user=self.user)
        self.staff_member.services_offered.add(self.test_service)

    def test_generate_random_id(self):
        id1 = generate_random_id()
        id2 = generate_random_id()
        self.assertNotEqual(id1, id2)

    # Test cases for get_timestamp
    def test_get_timestamp(self):
        ts = get_timestamp()
        self.assertIsNotNone(ts)
        self.assertIsInstance(ts, str)

    # Test cases for is_ajax
    def test_is_ajax_true(self):
        request = HttpRequest()
        request.META['HTTP_X_REQUESTED_WITH'] = 'XMLHttpRequest'
        self.assertTrue(is_ajax(request))

    def test_is_ajax_false(self):
        request = HttpRequest()
        self.assertFalse(is_ajax(request))

    # Test cases for get_available_slots
    def test_get_available_slots(self):
        date_str = datetime.date.today().strftime('%Y-%m-%d')
        date = convert_str_to_date(date_str)
        ar = AppointmentRequest.objects.create(
            date=date,
            start_time=datetime.time(9, 0),
            end_time=datetime.time(10, 0),
            service=self.test_service,
            staff_member=self.staff_member
        )
        appointment = Appointment.objects.create(appointment_request=ar, client=self.user)
        slots = get_available_slots(date, [appointment])
        self.assertIsInstance(slots, list)
        logging.info(slots)
        self.assertNotIn('09:00 AM', slots)

    def test_get_available_slots_with_config(self):
        date_str = datetime.date.today().strftime('%Y-%m-%d')
        date = convert_str_to_date(date_str)
        lead_time = datetime.time(8, 0)
        finish_time = datetime.time(17, 0)
        slot_duration = 30
        appointment_buffer_time = 2.0
        Config.objects.create(
            lead_time=lead_time,
            finish_time=finish_time,
            slot_duration=slot_duration,
            appointment_buffer_time=appointment_buffer_time
        )
        ar = AppointmentRequest.objects.create(
            date=date,
            start_time=datetime.time(9, 0),
            end_time=datetime.time(10, 0),
            service=self.test_service,
            staff_member=self.staff_member
        )
        appointment = Appointment.objects.create(appointment_request=ar, client=self.user)
        slots = get_available_slots(date, [appointment])
        self.assertIsInstance(slots, list)
        logging.info(slots)
        self.assertNotIn('09:00 AM', slots)
        # Additional assertions to verify that the slots are calculated correctly

    # Test cases for get_locale
    def test_get_locale_en(self):
        with self.settings(LANGUAGE_CODE='en'):
            self.assertEqual(get_locale(), 'en')

    def test_get_locale_en_us(self):
        with self.settings(LANGUAGE_CODE='en_US'):
            self.assertEqual(get_locale(), 'en')

    def test_get_locale_fr(self):
        # Set the local to French
        with self.settings(LANGUAGE_CODE='fr'):
            self.assertEqual(get_locale(), 'fr')

    def test_get_locale_fr_France(self):
        # Set the local to French
        with self.settings(LANGUAGE_CODE='fr_FR'):
            self.assertEqual(get_locale(), 'fr')

    def test_get_locale_others(self):
        with self.settings(LANGUAGE_CODE='de'):
            self.assertEqual(get_locale(), 'de')

    # Test cases for get_current_year
    def test_get_current_year(self):
        self.assertEqual(get_current_year(), datetime.datetime.now().year)

    # Test cases for get_timezone
    def test_get_timezone(self):
        self.assertIsNotNone(get_timezone())

    # Test cases for convert_str_to_date
    def test_convert_str_to_date_valid(self):
        date_str = '2023-07-31'
        date_obj = convert_str_to_date(date_str)
        self.assertEqual(date_obj, datetime.date(2023, 7, 31))

    def test_convert_str_to_date_invalid(self):
        date_str = 'invalid-date'
        with self.assertRaises(ValueError):
            convert_str_to_date(date_str)

    def test_get_website_name_no_config(self):
        website_name = get_website_name()
        self.assertEqual(website_name, APPOINTMENT_WEBSITE_NAME)

    def test_get_website_name_with_config(self):
        Config.objects.create(website_name="Test Website")
        website_name = get_website_name()
        self.assertEqual(website_name, "Test Website")

    def test_get_appointment_slot_duration_no_config(self):
        slot_duration = get_appointment_slot_duration()
        self.assertEqual(slot_duration, APPOINTMENT_SLOT_DURATION)

    def test_get_appointment_slot_duration_with_config(self):
        Config.objects.create(slot_duration=45)
        slot_duration = get_appointment_slot_duration()
        self.assertEqual(slot_duration, 45)

    # Test cases for get_appointment_lead_time
    def test_get_appointment_lead_time_no_config(self):
        lead_time = get_appointment_lead_time()
        self.assertEqual(lead_time, APPOINTMENT_LEAD_TIME)

    def test_get_appointment_lead_time_with_config(self):
        config_lead_time = datetime.time(hour=7, minute=30)
        Config.objects.create(lead_time=config_lead_time)
        lead_time = get_appointment_lead_time()
        self.assertEqual(lead_time, config_lead_time)

    # Test cases for get_appointment_finish_time
    def test_get_appointment_finish_time_no_config(self):
        finish_time = get_appointment_finish_time()
        self.assertEqual(finish_time, APPOINTMENT_FINISH_TIME)

    def test_get_appointment_finish_time_with_config(self):
        config_finish_time = datetime.time(hour=18, minute=30)
        Config.objects.create(finish_time=config_finish_time)
        finish_time = get_appointment_finish_time()
        self.assertEqual(finish_time, config_finish_time)

    # Test cases for get_appointment_buffer_time
    def test_get_appointment_buffer_time_no_config(self):
        buffer_time = get_appointment_buffer_time()
        self.assertEqual(buffer_time, APPOINTMENT_BUFFER_TIME)

    def test_get_appointment_buffer_time_with_config(self):
        config_buffer_time = 0.5  # 30 minutes
        Config.objects.create(appointment_buffer_time=config_buffer_time)
        buffer_time = get_appointment_buffer_time()
        self.assertEqual(buffer_time, config_buffer_time)

    # Test cases for get_finish_button_text
    def test_get_finish_button_text_free_service(self):
        service = Service(price=0)
        button_text = get_finish_button_text(service)
        self.assertEqual(button_text, _("Finish"))

    def test_get_finish_button_text_paid_service(self):
        with patch('appointment.services.APPOINTMENT_PAYMENT_URL', 'https://payment.com'):
            service = Service(price=100)
            button_text = get_finish_button_text(service)
            self.assertEqual(button_text, _("Pay Now"))

    # Test cases for get_user_model
    def test_get_client_model(self):
        client_model = get_user_model()
        self.assertEqual(client_model, apps.get_model(settings.AUTH_USER_MODEL))


class UtilityDateTimeTestCase(TestCase):
    def test_combine_date_and_time_success(self):
        """
        Test combining a date and time into a datetime object successfully.
        """
        date = datetime.date(2024, 2, 5)
        time = datetime.time(14, 30)
        expected_datetime = datetime.datetime(2024, 2, 5, 14, 30)
        combined_datetime = combine_date_and_time(date, time)
        self.assertEqual(combined_datetime, expected_datetime)

    def test_combine_date_and_time_with_timezone(self):
        """
        Test combining a date and time into a timezone-aware datetime object.
        """
        date = datetime.date(2024, 2, 5)
        time = datetime.time(14, 30, tzinfo=datetime.timezone.utc)
        combined_datetime = combine_date_and_time(date, time)
        self.assertTrue(timezone.is_aware(combined_datetime), "The datetime object is not timezone-aware.")
        self.assertEqual(combined_datetime.tzinfo, datetime.timezone.utc)

    def test_combine_date_and_time_with_invalid_date(self):
        """
        Test handling of invalid date input.
        """
        date = "2024-02-05"  # Incorrect type. It should be datetime.date
        time = datetime.time(14, 30)
        with self.assertRaises(TypeError, msg="Expected TypeError when combining with invalid date type"):
            combine_date_and_time(date, time)

    def test_combine_date_and_time_with_invalid_time(self):
        """
        Test handling of invalid time input.
        """
        date = datetime.date(2024, 2, 5)
        time = "14:30"  # Incorrect type. It should be datetime.time
        with self.assertRaises(TypeError, msg="Expected TypeError when combining with invalid time type"):
            combine_date_and_time(date, time)



# view_helpers.py
# Path: appointment/view_helpers.py

"""
Author: Adams Pierre David
Since: 2.0.0
"""

import uuid

from django.utils.translation import get_language, to_locale

from appointment.settings import APP_TIME_ZONE


def get_locale() -> str:
    """Get the current locale based on the user's language settings, without the country code.
    Used in the javascript files.
    Can't use the lang_country format because it is not supported.

    :return: The current locale as a string (language code only)
    """
    locale = to_locale(get_language())
    # Split the locale by '_' and take the first part (language code)
    return locale.split('_')[0]


def is_ajax(request):
    """Check if the request is an AJAX request.

    :param request: HttpRequest object
    :return: bool, True if the request is an AJAX request, False otherwise
    """
    return request.META.get('HTTP_X_REQUESTED_WITH') == 'XMLHttpRequest'


def generate_random_id() -> str:
    """Generate a random UUID and return it as a hexadecimal string.

    :return: The randomly generated UUID as a hex string
    """
    return uuid.uuid4().hex


def get_timezone_txt() -> str:
    # TODO: To be better implemented in the future
    """Get the current timezone as a string. To be used in the HTML template when the user is choosing the appointment's
    time.

    :return: The current timezone in a string format.
    """
    tmz = APP_TIME_ZONE
    timezone_map = {
        'UTC': 'Universal Time Coordinated (UTC)',
        'US/Eastern': 'Eastern Daylight Time (US & Canada)',
        'US/Central': 'Central Time (US & Canada)',
        'US/Mountain': 'Mountain Time (US & Canada)',
        'US/Pacific': 'Pacific Time (US & Canada)',
        'US/Alaska': 'Alaska Time (US & Canada)',
        'US/Hawaii': 'Hawaii Time (US & Canada)',
        'Europe/Paris': 'Paris Time (Europe)',
        'Europe/London': 'London Time (Europe)',
        'EDT': 'Eastern Daylight Time (US & Canada)',
    }
    return timezone_map.get(tmz, 'Universal Time Coordinated (UTC)')



# test_availability_slot.py
# Path: appointment/tests/test_availability_slot.py

from datetime import date, time, timedelta

from django.test import TestCase

from appointment.models import Appointment, AppointmentRequest
from appointment.tests.mixins.base_mixin import (
    AppointmentMixin, AppointmentRequestMixin, ConfigMixin, ServiceMixin, StaffMemberMixin, UserMixin)
from appointment.views import get_appointments_and_slots


class SlotAvailabilityTest(TestCase, UserMixin, ServiceMixin, StaffMemberMixin, AppointmentRequestMixin,
                           AppointmentMixin, ConfigMixin):
    def setUp(self):
        self.user = self.create_user_()
        self.service = self.create_service_(duration=timedelta(hours=2))
        self.staff_member = self.create_staff_member_(self.user, self.service)
        self.ar = self.create_appointment_request_(self.service, self.staff_member)
        self.appointment = self.create_appointment_(self.user, self.ar)
        self.config = self.create_config_(lead_time=time(11, 0), finish_time=time(15, 0), slot_duration=120)
        self.test_date = date.today() + timedelta(days=1)  # Use tomorrow's date for the tests

    def test_slot_availability_without_appointments(self):
        """Test if the available slots are correct when there are no appointments."""
        _, available_slots = get_appointments_and_slots(self.test_date, self.service)
        expected_slots = ['11:00 AM', '01:00 PM']
        self.assertEqual(available_slots, expected_slots)

    def test_slot_availability_with_first_slot_booked(self):
        """Available slots (total 2) should be one when the first slot is booked."""
        ar = AppointmentRequest.objects.create(date=self.test_date, start_time=time(11, 0), end_time=time(13, 0),
                                               service=self.service, staff_member=self.staff_member)
        Appointment.objects.create(client=self.user, appointment_request=ar)
        _, available_slots = get_appointments_and_slots(self.test_date, self.service)
        expected_slots = ['01:00 PM']
        self.assertEqual(available_slots, expected_slots)

    def test_slot_availability_with_second_slot_booked(self):
        """Available slots (total 2) should be one when the second slot is booked."""
        ar = AppointmentRequest.objects.create(date=self.test_date, start_time=time(13, 0), end_time=time(15, 0),
                                               service=self.service, staff_member=self.staff_member)
        Appointment.objects.create(client=self.user, appointment_request=ar)
        _, available_slots = get_appointments_and_slots(self.test_date, self.service)
        expected_slots = ['11:00 AM']
        self.assertEqual(available_slots, expected_slots)

    def test_slot_availability_with_both_slots_booked(self):
        """Available slots (total 2) should be zero when both slots are booked."""
        ar1 = AppointmentRequest.objects.create(date=self.test_date, start_time=time(11, 0), end_time=time(13, 0),
                                                service=self.service, staff_member=self.staff_member)
        ar2 = AppointmentRequest.objects.create(date=self.test_date, start_time=time(13, 0), end_time=time(15, 0),
                                                service=self.service, staff_member=self.staff_member)
        Appointment.objects.create(client=self.user, appointment_request=ar1)
        Appointment.objects.create(client=self.user, appointment_request=ar2)
        _, available_slots = get_appointments_and_slots(self.test_date, self.service)
        expected_slots = []
        self.assertEqual(available_slots, expected_slots)



# permissions.py
# Path: appointment/utils/permissions.py

"""
Author: Adams Pierre David
Since: 2.0.0
"""


def check_entity_ownership(user, entity):
    """Check if a user owns an entity."""
    return entity.is_owner(user.pk) or user.is_superuser


def check_extensive_permissions(staff_user_id, user, entity):
    """Check if a user has permissions to make changes based on staff_user_id and entity ownership."""
    if (staff_user_id and int(staff_user_id) == user.pk) or user.is_superuser:
        if entity:
            return check_entity_ownership(user, entity)
        else:
            return True
    return False


def check_permissions(staff_user_id, user):
    """Check if a user has permissions to add based on staff_user_id."""
    if (staff_user_id and int(staff_user_id) == user.pk) or user.is_superuser:
        return True
    return False


def has_permission_to_delete_appointment(user, appointment):
    """
    Check if the user has permission to delete the given appointment.
    Returns True if the user has permission, False otherwise.
    """
    return check_extensive_permissions(appointment.get_staff_member().user_id, user, appointment)



import string

from django.test import TestCase

from appointment.models import EmailVerificationCode
from appointment.tests.mixins.base_mixin import UserMixin


class EmailVerificationCodeModelTestCase(TestCase, UserMixin):
    def setUp(self):
        self.user = self.create_user_()
        self.code = EmailVerificationCode.generate_code(self.user)

    def test_code_creation(self):
        """Test if a verification code can be generated."""
        verification_code = EmailVerificationCode.objects.get(user=self.user)
        self.assertIsNotNone(verification_code)
        self.assertEqual(verification_code.code, self.code)

    def test_code_length(self):
        """Test that the code is six characters long."""
        self.assertEqual(len(self.code), 6)

    def test_code_content(self):
        """Test that the code only contains uppercase letters and digits."""
        valid_characters = set(string.ascii_uppercase + string.digits)
        self.assertTrue(all(char in valid_characters for char in self.code))

    def test_code_str_representation(self):
        """Test that the string representation of a verification code is correct."""
        verification_code = EmailVerificationCode.objects.get(user=self.user)
        self.assertEqual(str(verification_code), self.code)

    def test_created_at(self):
        """Test if the created_at field is set when a code is generated."""
        verification_code = EmailVerificationCode.objects.get(user=self.user)
        self.assertIsNotNone(verification_code.created_at)

    def test_updated_at(self):
        """Test if the updated_at field is set when a code is generated."""
        verification_code = EmailVerificationCode.objects.get(user=self.user)
        self.assertIsNotNone(verification_code.updated_at)

    def test_multiple_codes_for_user(self):
        """
        Test if multiple verification codes can be generated for a user.
        This should ideally create a new code, but the old one will still exist.
        """
        new_code = EmailVerificationCode.generate_code(self.user)
        self.assertNotEqual(self.code, new_code)

    def test_code_verification_match(self):
        """The check_code method returns True when the code matches."""
        code = EmailVerificationCode.objects.get(user=self.user)
        self.assertTrue(code.check_code(self.code))

    def test_code_verification_mismatch(self):
        """The check_code method returns False when the code does not match."""
        mismatched_code = "ABCDEF"
        code = EmailVerificationCode.objects.get(user=self.user)
        self.assertFalse(code.check_code(mismatched_code))



from datetime import datetime, time, timedelta

from django.core.exceptions import ValidationError
from django.db import IntegrityError
from django.utils import timezone

from appointment.models import Appointment, DayOff, WorkingHours
from appointment.tests.base.base_test import BaseTest
from appointment.utils.date_time import get_weekday_num


class AppointmentModelTestCase(BaseTest):
    def setUp(self):
        super().setUp()
        self.ar = self.create_appt_request_for_sm1()
        self.appointment = self.create_appointment_for_user1(appointment_request=self.ar)

    # Test appointment creation
    def test_appointment_creation(self):
        """Test if an appointment can be created."""
        appointment = Appointment.objects.get(appointment_request=self.ar)
        self.assertIsNotNone(appointment)
        self.assertEqual(appointment.client, self.client1)
        self.assertEqual(appointment.phone, "1234567890")
        self.assertEqual(appointment.address, "Some City, Some State")

    # Test str representation
    def test_str_representation(self):
        """Test if an appointment's string representation is correct."""
        expected_str = f"{self.client1} - {self.ar.start_time.strftime('%Y-%m-%d %H:%M')} to " \
                       f"{self.ar.end_time.strftime('%Y-%m-%d %H:%M')}"
        self.assertEqual(str(self.appointment), expected_str)

    # Test start time
    def test_get_start_time(self):
        """Test if an appointment's start time is correct."""
        expected_start_time = datetime.combine(self.ar.date,
                                               self.ar.start_time)
        self.assertEqual(self.appointment.get_start_time(), expected_start_time)

    # Test end time
    def test_get_end_time(self):
        """Test if an appointment's end time is correct."""
        expected_end_time = datetime.combine(self.ar.date,
                                             self.ar.end_time)
        self.assertEqual(self.appointment.get_end_time(), expected_end_time)

    # Test service name retrieval
    def test_get_service_name(self):
        """Test if an appointment's service name is correct."""
        self.assertEqual(self.appointment.get_service_name(), "Test Service")

    # Test service price retrieval
    def test_get_service_price(self):
        """Test if an appointment's service price is correct."""
        self.assertEqual(self.appointment.get_service_price(), 100)

    # Test phone retrieval
    def test_get_phone(self):
        """Test if an appointment's phone number is correct."""
        self.assertEqual(self.appointment.phone, "1234567890")

    # Test address retrieval
    def test_get_address(self):
        """Test if an appointment's address is correct."""
        self.assertEqual(self.appointment.address, "Some City, Some State")

    # Test reminder retrieval
    def test_get_want_reminder(self):
        """Test if an appointment's reminder status is correct."""
        self.assertFalse(self.appointment.want_reminder)

    # Test additional info retrieval
    def test_get_additional_info(self):
        """Test if an appointment's additional info is correct."""
        self.assertIsNone(self.appointment.additional_info)

    # Test paid status retrieval
    def test_is_paid(self):
        """Test if an appointment's paid status is correct."""
        self.assertFalse(self.appointment.is_paid())

    def test_is_paid_text(self):
        """Test if an appointment's paid status is correct."""
        self.assertEqual(self.appointment.is_paid_text(), "No")

    # Test appointment amount to pay
    def test_get_appointment_amount_to_pay(self):
        """Test if an appointment's amount to pay is correct."""
        self.assertEqual(self.appointment.get_appointment_amount_to_pay(), 100)

    # Test appointment currency retrieval
    def test_get_appointment_currency(self):
        """Test if an appointment's currency is correct."""
        self.assertEqual(self.appointment.get_appointment_currency(), "USD")

    # Test appointment ID request retrieval
    def test_get_appointment_id_request(self):
        """Test if an appointment's ID request is correct."""
        self.assertIsNotNone(self.appointment.get_appointment_id_request())

    # Test created at retrieval
    def test_created_at(self):
        """Test if an appointment's created at date is correct."""
        self.assertIsNotNone(self.appointment.created_at)

    # Test updated at retrieval
    def test_updated_at(self):
        """Test if an appointment's updated at date is correct."""
        self.assertIsNotNone(self.appointment.updated_at)

    # Test paid status setting
    def test_set_appointment_paid_status(self):
        """Test if an appointment's paid status can be set."""
        self.appointment.set_appointment_paid_status(True)
        self.assertTrue(self.appointment.is_paid())
        self.appointment.set_appointment_paid_status(False)
        self.assertFalse(self.appointment.is_paid())

    # Test invalid phone number
    def test_invalid_phone(self):
        """Test that an appointment cannot be created with an invalid phone number."""
        self.appointment.phone = "1234"  # Invalid phone number
        with self.assertRaises(ValidationError):
            self.appointment.full_clean()

    # Test service down payment retrieval
    def test_get_service_down_payment(self):
        """Test if an appointment's service down payment is correct."""
        self.assertEqual(self.appointment.get_service_down_payment(), self.service1.get_down_payment())

    # Test service description retrieval
    def test_service_description(self):
        """Test if an appointment's service description is correct."""
        self.assertEqual(self.appointment.get_service_description(), self.service1.description)

    # Test appointment date retrieval
    def test_get_appointment_date(self):
        """Test if an appointment's date is correct."""
        self.assertEqual(self.appointment.get_appointment_date(), self.ar.date)

    # Test save function with down payment type
    def test_save_with_down_payment(self):
        """Test if an appointment can be saved with a down payment."""
        self.ar.payment_type = 'down'
        self.ar.save()
        self.appointment.save()
        self.assertEqual(self.appointment.get_service_down_payment(), self.service1.get_down_payment())

    def test_appointment_without_appointment_request(self):
        """Test that an appointment cannot be created without an appointment request."""
        with self.assertRaises(ValidationError):  # Assuming model validation prevents this
            Appointment.objects.create(client=self.client1)

    def test_appointment_without_client(self):
        """Test that an appointment cannot be created without a client."""
        with self.assertRaises(IntegrityError):  # Assuming model validation prevents this
            Appointment.objects.create(appointment_request=self.ar)

    def test_appointment_amount_to_pay_calculation(self):
        """
        Test if an appointment's amount_to_pay field is correctly calculated based on the associated AppointmentRequest.
        """
        self.assertEqual(self.appointment.get_appointment_amount_to_pay(), self.ar.get_service_price())

    def test_update_appointment_paid_status(self):
        """Simulate appointment's paid status being updated."""
        self.appointment.set_appointment_paid_status(True)
        self.assertTrue(self.appointment.is_paid())
        self.appointment.set_appointment_paid_status(False)
        self.assertFalse(self.appointment.is_paid())

    def test_appointment_rescheduling(self):
        """Simulate appointment rescheduling by changing the appointment date and times."""
        new_date = self.ar.date + timedelta(days=1)
        new_start_time = time(10, 0)
        new_end_time = time(11, 0)
        self.ar.date = new_date
        self.ar.start_time = new_start_time
        self.ar.end_time = new_end_time
        self.ar.save()

        self.assertEqual(self.appointment.get_date(), new_date)
        self.assertEqual(self.appointment.get_start_time().time(), new_start_time)
        self.assertEqual(self.appointment.get_end_time().time(), new_end_time)

    def test_create_appointment_without_required_fields(self):
        """Test that an appointment cannot be created without the required fields."""
        with self.assertRaises(ValidationError):
            Appointment.objects.create()

    def test_get_service_duration(self):
        """Test if an appointment's service duration is correct."""
        self.assertEqual(self.appointment.get_service_duration(), "1 hour")

    def test_appt_to_dict(self):
        response = {
            'id': 1,
            'client_name': 'Client1',
            'client_email': 'client1@gmail.com',
            'start_time': '1900-01-01 09:00',
            'end_time': '1900-01-01 10:00',
            'service_name': 'Test Service',
            'address': 'Some City, Some State',
            'want_reminder': False,
            'additional_info': None,
            'paid': False,
            'amount_to_pay': 100,
        }
        actual_response = self.appointment.to_dict()
        actual_response.pop('id_request', None)
        self.assertEqual(actual_response, response)

    def test_get_staff_member_name_with_staff_member(self):
        """Test if you get_staff_member_name method returns the correct name when a staff member is associated."""
        expected_name = self.staff_member1.get_staff_member_name()
        actual_name = self.appointment.get_staff_member_name()
        self.assertEqual(actual_name, expected_name)

    def test_get_staff_member_name_without_staff_member(self):
        """Test if you get_staff_member_name method returns an empty string when no staff member is associated."""
        self.appointment.appointment_request.staff_member = None
        self.appointment.appointment_request.save()
        self.assertEqual(self.appointment.get_staff_member_name(), "")

    def test_get_service_img_url_no_image(self):
        """Service should handle cases where no image is provided gracefully."""
        self.assertEqual(self.appointment.get_service_img_url(), "")


class AppointmentValidDateTestCase(BaseTest):
    def setUp(self):
        super().setUp()
        self.weekday = "Monday"  # Example weekday
        self.weekday_num = get_weekday_num(self.weekday)
        WorkingHours.objects.create(staff_member=self.staff_member1, day_of_week=self.weekday_num,
                                    start_time=time(9, 0),
                                    end_time=time(17, 0))
        self.appt_date = timezone.now().date() + timedelta(days=(self.weekday_num - timezone.now().weekday()) % 7)
        self.start_time = timezone.now().replace(hour=10, minute=0, second=0, microsecond=0)
        self.current_appointment_id = None

    def test_staff_member_works_on_given_day(self):
        is_valid, message = Appointment.is_valid_date(self.appt_date, self.start_time, self.staff_member1,
                                                      self.current_appointment_id, self.weekday)
        self.assertTrue(is_valid)

    def test_staff_member_does_not_work_on_given_day(self):
        non_working_day = "Sunday"
        non_working_day_num = get_weekday_num(non_working_day)
        appt_date = self.appt_date + timedelta(days=(non_working_day_num - self.weekday_num) % 7)
        is_valid, message = Appointment.is_valid_date(appt_date, self.start_time, self.staff_member1,
                                                      self.current_appointment_id, non_working_day)
        self.assertFalse(is_valid)
        self.assertIn("does not work on this day", message)

    def test_start_time_outside_working_hours(self):
        early_start_time = timezone.now().replace(hour=8, minute=0)  # Before working hours
        is_valid, message = Appointment.is_valid_date(self.appt_date, early_start_time, self.staff_member1,
                                                      self.current_appointment_id, self.weekday)
        self.assertFalse(is_valid)
        self.assertIn("outside of", message)

    def test_staff_member_has_day_off(self):
        DayOff.objects.create(staff_member=self.staff_member1, start_date=self.appt_date, end_date=self.appt_date)
        is_valid, message = Appointment.is_valid_date(self.appt_date, self.start_time, self.staff_member1,
                                                      self.current_appointment_id, self.weekday)
        self.assertFalse(is_valid)
        self.assertIn("has a day off on this date", message)



from datetime import timedelta

from django.core.exceptions import ValidationError
from django.utils import timezone

from appointment.models import AppointmentRescheduleHistory
from appointment.tests.base.base_test import BaseTest


class AppointmentRescheduleHistoryTestCase(BaseTest):
    def setUp(self):
        super().setUp()
        self.appointment_request = self.create_appt_request_for_sm1()

    def test_successful_creation(self):
        reschedule_history = AppointmentRescheduleHistory.objects.create(
            appointment_request=self.appointment_request,
            date=timezone.now().date() + timedelta(days=1),  # Future date
            start_time=timezone.now().time(),
            end_time=(timezone.now() + timedelta(hours=1)).time(),
            staff_member=self.staff_member1,
            reason_for_rescheduling="Client request",
            reschedule_status='pending'
        )
        self.assertIsNotNone(reschedule_history.id_request)  # Auto-generated id_request
        self.assertTrue(reschedule_history.still_valid())

    def test_date_in_past_validation(self):
        with self.assertRaises(ValidationError):
            AppointmentRescheduleHistory.objects.create(
                appointment_request=self.appointment_request,
                date=timezone.now().date() - timedelta(days=1),  # Past date
                start_time=timezone.now().time(),
                end_time=(timezone.now() + timedelta(hours=1)).time(),
                staff_member=self.staff_member1
            )

    def test_invalid_date_validation(self):
        with self.assertRaises(TypeError):
            AppointmentRescheduleHistory.objects.create(
                appointment_request=self.appointment_request,
                date="invalid-date",  # Invalid date format
                start_time=timezone.now().time(),
                end_time=(timezone.now() + timedelta(hours=1)).time(),
                staff_member=self.staff_member1
            )

    def test_still_valid(self):
        reschedule_history = AppointmentRescheduleHistory.objects.create(
            appointment_request=self.appointment_request,
            date=timezone.now().date() + timedelta(days=1),
            start_time=timezone.now().time(),
            end_time=(timezone.now() + timedelta(hours=1)).time(),
            staff_member=self.staff_member1,
            reason_for_rescheduling="Client request",
            reschedule_status='pending'
        )
        # Directly test the still_valid method
        self.assertTrue(reschedule_history.still_valid())

        # Simulate passages of time beyond the validity window
        reschedule_history.created_at -= timedelta(minutes=6)
        reschedule_history.save()
        self.assertFalse(reschedule_history.still_valid())



# test_date_time.py
# Path: appointment/tests/utils/test_date_time.py

import datetime
from unittest.mock import Mock, patch

from django.test import TestCase

from appointment.settings import APP_TIME_ZONE
from appointment.utils.date_time import (
    combine_date_and_time, convert_12_hour_time_to_24_hour_time, convert_24_hour_time_to_12_hour_time,
    convert_minutes_in_human_readable_format, convert_str_to_date,
    convert_str_to_time, get_ar_end_time, get_current_year, get_timestamp, get_timezone, get_weekday_num,
    time_difference
)


class Convert12HourTo24HourTimeTests(TestCase):
    def test_valid_basic_conversion(self):
        """Test basic 12-hour to 24-hour conversions."""
        self.assertEqual(convert_12_hour_time_to_24_hour_time("01:10 AM"), "01:10:00")
        self.assertEqual(convert_12_hour_time_to_24_hour_time("01:20 PM"), "13:20:00")

    def test_convert_midnight_and_noon(self):
        """Test conversion of midnight and noon times."""
        self.assertEqual(convert_12_hour_time_to_24_hour_time("12:00 AM"), "00:00:00")
        self.assertEqual(convert_12_hour_time_to_24_hour_time("12:00 PM"), "12:00:00")

    def test_boundary_times(self):
        """Test conversion of boundary times."""
        self.assertEqual(convert_12_hour_time_to_24_hour_time("12:00 AM"), "00:00:00")
        self.assertEqual(convert_12_hour_time_to_24_hour_time("11:59 PM"), "23:59:00")

    def test_datetime_and_time_objects(self):
        """Test conversion using datetime and time objects."""
        dt_obj = datetime.datetime(2023, 1, 1, 14, 30)
        time_obj = datetime.time(14, 30)
        self.assertEqual(convert_12_hour_time_to_24_hour_time(dt_obj), "14:30:00")
        self.assertEqual(convert_12_hour_time_to_24_hour_time(time_obj), "14:30:00")

    def test_case_insensitivity_and_whitespace(self):
        """Test conversion handling of different case formats and white-space."""
        self.assertEqual(convert_12_hour_time_to_24_hour_time(" 12:00 am "), "00:00:00")
        self.assertEqual(convert_12_hour_time_to_24_hour_time("01:00 pM "), "13:00:00")

    def test_out_of_bounds_values(self):
        """Test conversion with out-of-bounds values."""
        with self.assertRaises(ValueError):
            convert_12_hour_time_to_24_hour_time("13:00 PM")
        with self.assertRaises(ValueError):
            convert_12_hour_time_to_24_hour_time("12:60 AM")

    def test_invalid_datatypes(self):
        """Test conversion with invalid data types."""
        with self.assertRaises(ValueError):
            convert_12_hour_time_to_24_hour_time(["12:00 AM"])
        with self.assertRaises(ValueError):
            convert_12_hour_time_to_24_hour_time({"time": "12:00 AM"})

    def test_invalid_inputs(self):
        """Test conversion with various invalid inputs."""
        with self.assertRaises(ValueError):
            convert_12_hour_time_to_24_hour_time("25:00 AM")
        with self.assertRaises(ValueError):
            convert_12_hour_time_to_24_hour_time("01:00")
        with self.assertRaises(ValueError):
            convert_12_hour_time_to_24_hour_time("Random String")
        with self.assertRaises(ValueError):
            convert_12_hour_time_to_24_hour_time("01:60 AM")


class Convert24HourTimeTo12HourTimeTests(TestCase):

    def test_valid_24_hour_time_strings(self):
        self.assertEqual(convert_24_hour_time_to_12_hour_time("13:00"), "01:00 PM")
        self.assertEqual(convert_24_hour_time_to_12_hour_time("00:00"), "12:00 AM")
        self.assertEqual(convert_24_hour_time_to_12_hour_time("23:59"), "11:59 PM")
        self.assertEqual(convert_24_hour_time_to_12_hour_time("12:00"), "12:00 PM")
        self.assertEqual(convert_24_hour_time_to_12_hour_time("01:00"), "01:00 AM")

    def test_valid_24_hour_time_with_seconds(self):
        self.assertEqual(convert_24_hour_time_to_12_hour_time("13:00:01"), "01:00:01 PM")
        self.assertEqual(convert_24_hour_time_to_12_hour_time("00:00:59"), "12:00:59 AM")

    def test_datetime_time_object_input(self):
        time_input = datetime.time(13, 15)
        self.assertEqual(convert_24_hour_time_to_12_hour_time(time_input), "01:15 PM")
        time_input = datetime.time(0, 0)
        self.assertEqual(convert_24_hour_time_to_12_hour_time(time_input), "12:00 AM")

    def test_invalid_time_strings(self):
        with self.assertRaises(ValueError):
            convert_24_hour_time_to_12_hour_time("25:00")
        with self.assertRaises(ValueError):
            convert_24_hour_time_to_12_hour_time("-01:00")
        with self.assertRaises(ValueError):
            convert_24_hour_time_to_12_hour_time("13:60")
        with self.assertRaises(ValueError):
            convert_24_hour_time_to_12_hour_time("invalid")

    def test_incorrect_format(self):
        with self.assertRaises(ValueError):
            convert_24_hour_time_to_12_hour_time("1 PM")
        with self.assertRaises(ValueError):
            convert_24_hour_time_to_12_hour_time("13 PM")
        with self.assertRaises(ValueError):
            convert_24_hour_time_to_12_hour_time("24:00")

    def test_edge_cases(self):
        self.assertEqual(convert_24_hour_time_to_12_hour_time("12:00"), "12:00 PM")
        self.assertEqual(convert_24_hour_time_to_12_hour_time("00:00"), "12:00 AM")
        self.assertEqual(convert_24_hour_time_to_12_hour_time("11:59"), "11:59 AM")
        self.assertEqual(convert_24_hour_time_to_12_hour_time("23:59"), "11:59 PM")


class ConvertMinutesInHumanReadableFormatTests(TestCase):
    def test_valid_basic_conversions(self):
        """Test basic conversions"""
        self.assertEqual(convert_minutes_in_human_readable_format(30), "30 minutes")
        self.assertEqual(convert_minutes_in_human_readable_format(90), "1 hour and 30 minutes")

    def test_edge_cases(self):
        """Test edgy cases"""
        self.assertEqual(convert_minutes_in_human_readable_format(59), "59 minutes")
        self.assertEqual(convert_minutes_in_human_readable_format(60), "1 hour")
        self.assertEqual(convert_minutes_in_human_readable_format(1439), "23 hours and 59 minutes")
        # '1440' minutes is the total number of minutes in a day, hence it should convert to "1 day".
        self.assertEqual(convert_minutes_in_human_readable_format(1440), "1 day")

    def test_valid_combinations(self):
        """Test various combinations"""
        self.assertEqual(convert_minutes_in_human_readable_format(1441), "1 day and 1 minute")
        self.assertEqual(convert_minutes_in_human_readable_format(1500), "1 day and 1 hour")
        self.assertEqual(convert_minutes_in_human_readable_format(1560), "1 day and 2 hours")
        self.assertEqual(convert_minutes_in_human_readable_format(1501), "1 day, 1 hour and 1 minute")
        self.assertEqual(convert_minutes_in_human_readable_format(1562), "1 day, 2 hours and 2 minutes")
        self.assertEqual(convert_minutes_in_human_readable_format(808), "13 hours and 28 minutes")

    def test_non_positive_values(self):
        """Test that non-positive values are handled correctly"""
        self.assertEqual(convert_minutes_in_human_readable_format(0), "Not set.")
        # Expectation for negative values might depend on desired behavior, just an example below
        with self.assertRaises(ValueError):
            convert_minutes_in_human_readable_format(-5)

    def test_float_values(self):
        """Test float values which should be correctly rounded down"""
        self.assertEqual(convert_minutes_in_human_readable_format(2.5), "2 minutes")
        self.assertEqual(convert_minutes_in_human_readable_format(2.9), "2 minutes")

    def test_invalid_inputs(self):
        """Test invalid inputs which should raise an error"""
        with self.assertRaises(TypeError):
            convert_minutes_in_human_readable_format("30 minutes")
        with self.assertRaises(TypeError):
            convert_minutes_in_human_readable_format(["30"])
        with self.assertRaises(TypeError):
            convert_minutes_in_human_readable_format({"minutes": 30})
        with self.assertRaises(TypeError):
            convert_minutes_in_human_readable_format(None)


class ConvertStrToDateTests(TestCase):

    def test_valid_date_strings_with_hyphen_separator(self):
        """Test valid date with hyphen separator works correctly"""
        self.assertEqual(convert_str_to_date("2023-12-31"), datetime.date(2023, 12, 31))
        self.assertEqual(convert_str_to_date("2020-02-29"), datetime.date(2020, 2, 29))  # Leap year
        self.assertEqual(convert_str_to_date("2021-02-28"), datetime.date(2021, 2, 28))

    def test_valid_date_strings_with_slash_separator(self):
        """Test valid date with slash separator works correctly"""
        self.assertEqual(convert_str_to_date("2021/01/01"), datetime.date(2021, 1, 1))
        self.assertEqual(convert_str_to_date("2023/12/31"), datetime.date(2023, 12, 31))
        self.assertEqual(convert_str_to_date("2023.12.31"), datetime.date(2023, 12, 31))

    def test_invalid_date_formats(self):
        """The date format "MM-DD-YYY" & "DD/MM/YYYY" are not supported, hence it should raise an error."""
        with self.assertRaises(ValueError):
            convert_str_to_date("12-31-2023")
        with self.assertRaises(ValueError):
            convert_str_to_date("31/12/2023")

    def test_nonexistent_dates(self):
        """Test nonexistent dates"""
        with self.assertRaises(ValueError):
            convert_str_to_date("2023-02-30")
        with self.assertRaises(ValueError):
            convert_str_to_date("2021-02-29")  # Not a leap year

    def test_other_invalid_inputs(self):
        """Test other invalid inputs"""
        with self.assertRaises(ValueError):
            convert_str_to_date("")
        with self.assertRaises(ValueError):
            convert_str_to_date("RandomString")
        with self.assertRaises(ValueError):
            convert_str_to_date("0000-00-00")


class ConvertStrToTimeTests(TestCase):
    def test_convert_12h_format_str_to_time(self):
        """Test valid time strings"""
        # These tests check if a 12-hour time format string converts correctly.
        self.assertEqual(convert_str_to_time("10:00 AM"), datetime.time(10, 0))
        self.assertEqual(convert_str_to_time("12:00 PM"), datetime.time(12, 0))
        self.assertEqual(convert_str_to_time("01:30 PM"), datetime.time(13, 30))

    def test_convert_24h_format_str_to_time(self):
        """Test if a 24-hour time format string converts correctly."""
        # These tests check if a 24-hour time format string converts correctly.
        self.assertEqual(convert_str_to_time("10:00:00"), datetime.time(10, 0))
        self.assertEqual(convert_str_to_time("12:00:00"), datetime.time(12, 0))
        self.assertEqual(convert_str_to_time("13:30:00"), datetime.time(13, 30))

    def test_case_insensitivity_and_whitespace(self):
        """Test conversion handling of different case formats and white-space."""
        self.assertEqual(convert_str_to_time(" 12:00 am "), datetime.time(0, 0))
        self.assertEqual(convert_str_to_time("01:00 pM "), datetime.time(13, 0))
        self.assertEqual(convert_str_to_time(" 13:00:00 "), datetime.time(13, 0))

    def test_convert_str_to_time_invalid(self):
        """Test invalid time strings"""
        with self.assertRaises(ValueError):
            convert_str_to_time("")
        with self.assertRaises(ValueError):
            convert_str_to_time("13:00 PM")
        with self.assertRaises(ValueError):
            convert_str_to_time("25:00 AM")
        with self.assertRaises(ValueError):
            convert_str_to_time("25:00:00")
        with self.assertRaises(ValueError):
            convert_str_to_time("10:60 AM")
        with self.assertRaises(ValueError):
            convert_str_to_time("10:60:00")
        with self.assertRaises(ValueError):
            convert_str_to_time("Random String")


class GetAppointmentRequestEndTimeTests(TestCase):
    def test_get_ar_end_time_with_valid_inputs(self):
        """Test positive cases"""
        self.assertEqual(get_ar_end_time("10:00:00", 60), datetime.time(11, 0))
        self.assertEqual(get_ar_end_time(datetime.time(10, 0), 120), datetime.time(12, 0))
        self.assertEqual(get_ar_end_time(datetime.time(10, 0), datetime.timedelta(hours=2)), datetime.time(12, 0))

    def test_negative_duration(self):
        """Test negative duration"""
        with self.assertRaises(ValueError):
            get_ar_end_time("10:00:00", -60)

    def test_invalid_start_time_format(self):
        """Test invalid start time format"""
        with self.assertRaises(ValueError):
            get_ar_end_time("25:00:00", 60)

    def test_invalid_duration_format(self):
        """Test invalid duration format"""
        with self.assertRaises(TypeError):
            get_ar_end_time("10:00:00", "60")

    def test_end_time_past_midnight(self):
        """Test end time past midnight"""
        # If the end time goes past midnight, it should wrap around to the next day,
        # hence "23:30:00" + 60 minutes = "00:30:00".
        self.assertEqual(get_ar_end_time("23:30:00", 60), datetime.time(0, 30))

    def test_invalid_start_time_type(self):
        """Test that an invalid start_time type raises a TypeError."""
        with self.assertRaises(TypeError):
            get_ar_end_time([], 60)  # Passing a list instead of a datetime.time object or string
        with self.assertRaises(TypeError):
            get_ar_end_time(12345, 30)  # Passing an integer
        with self.assertRaises(TypeError):
            get_ar_end_time(None, 30)  # Passing None


class TimeDifferenceTests(TestCase):

    def test_valid_difference_with_time_objects(self):
        """Test difference between two time objects"""
        time1 = datetime.time(10, 0)
        time2 = datetime.time(11, 0)
        difference = time_difference(time1, time2)
        self.assertEqual(difference, datetime.timedelta(hours=1))

    def test_valid_difference_with_datetime_objects(self):
        """Test difference between two datetime objects"""
        datetime1 = datetime.datetime(2023, 1, 1, 10, 0)
        datetime2 = datetime.datetime(2023, 1, 1, 11, 0)
        difference = time_difference(datetime1, datetime2)
        self.assertEqual(difference, datetime.timedelta(hours=1))

    def test_negative_difference_with_time_objects(self):
        """Two time objects cannot have a negative difference"""
        time1 = datetime.time(10, 0)
        time2 = datetime.time(11, 0)
        with self.assertRaises(ValueError):
            time_difference(time2, time1)

    def test_negative_difference_with_datetime_objects(self):
        """Two datetime objects cannot have a negative difference"""
        datetime1 = datetime.datetime(2023, 1, 1, 10, 0)
        datetime2 = datetime.datetime(2023, 1, 1, 11, 0)
        with self.assertRaises(ValueError):
            time_difference(datetime2, datetime1)

    def test_mismatched_input_types(self):
        """Test that providing one 'datetime.time' and one datetime.datetime raises a ValueError."""
        time_obj = datetime.time(10, 0)
        datetime_obj = datetime.datetime(2023, 1, 1, 11, 0)

        with self.assertRaises(ValueError) as context:
            time_difference(time_obj, datetime_obj)

        self.assertEqual(str(context.exception),
                         "Both inputs should be of the same type, either datetime.time or datetime.datetime")

        # Test the reverse case as well for completeness
        with self.assertRaises(ValueError) as context:
            time_difference(datetime_obj, time_obj)

        self.assertEqual(str(context.exception),
                         "Both inputs should be of the same type, either datetime.time or datetime.datetime")


class CombineDateAndTimeTests(TestCase):
    def test_combine_valid_date_and_time(self):
        """Test combining a valid date and time."""
        date = datetime.date(2023, 1, 1)
        time = datetime.time(12, 30)
        expected_datetime = datetime.datetime(2023, 1, 1, 12, 30)
        result = combine_date_and_time(date, time)
        self.assertEqual(result, expected_datetime)

    def test_combine_with_midnight(self):
        """Test combining a date with a midnight time."""
        date = datetime.date(2023, 1, 1)
        time = datetime.time(0, 0)
        expected_datetime = datetime.datetime(2023, 1, 1, 0, 0)
        result = combine_date_and_time(date, time)
        self.assertEqual(result, expected_datetime)

    def test_combine_with_leap_year_date(self):
        """Test combining a leap year date and time."""
        date = datetime.date(2024, 2, 29)
        time = datetime.time(23, 59)
        expected_datetime = datetime.datetime(2024, 2, 29, 23, 59)
        result = combine_date_and_time(date, time)
        self.assertEqual(result, expected_datetime)

    def test_combine_with_end_of_day(self):
        """Test combining a date with 'end of day time'."""
        date = datetime.date(2023, 1, 1)
        time = datetime.time(23, 59, 59)
        expected_datetime = datetime.datetime(2023, 1, 1, 23, 59, 59)
        result = combine_date_and_time(date, time)
        self.assertEqual(result, expected_datetime)

    def test_combine_with_microseconds(self):
        """Test combining a date and time with microseconds."""
        date = datetime.date(2023, 1, 1)
        time = datetime.time(12, 30, 15, 123456)
        expected_datetime = datetime.datetime(2023, 1, 1, 12, 30, 15, 123456)
        result = combine_date_and_time(date, time)
        self.assertEqual(result, expected_datetime)


class TimestampTests(TestCase):
    @patch('appointment.utils.date_time.timezone.now')
    def test_get_timestamp(self, mock_now):
        """Test get_timestamp function"""
        mock_datetime = Mock()
        mock_datetime.timestamp.return_value = 1612345678.1234  # Sample timestamp with decimal
        mock_now.return_value = mock_datetime

        self.assertEqual(get_timestamp(), "16123456781234")


class GeneralDateTimeTests(TestCase):
    def test_get_timezone(self):
        """Test get_timezone function"""
        self.assertEqual(get_timezone(), APP_TIME_ZONE)

    def test_get_current_year(self):
        """Test get_current_year function"""
        self.assertEqual(get_current_year(), datetime.datetime.now().year)

    def test_get_current_year_mocked(self):
        """Test get_current_year function with a mocked year."""
        with patch('appointment.utils.date_time.datetime.datetime') as mock_date:
            mock_date.now.return_value.year = 1999  # Setting year attribute of the mock object
            self.assertEqual(get_current_year(), 1999)

    @patch('appointment.utils.date_time.APP_TIME_ZONE', new="Asia/Tokyo")
    def test_get_timezone_with_modified_setting(self):
        """Test get_timezone function with a modified timezone setting."""
        self.assertEqual(get_timezone(), "Asia/Tokyo")

    def test_get_weekday_num(self):
        """Test get_weekday_num function with valid input"""
        self.assertEqual(get_weekday_num("Monday"), 1)
        self.assertEqual(get_weekday_num("Sunday"), 0)

    def test_invalid_get_weekday_num(self):
        """Test get_weekday_num function with invalid input which should return -1"""
        self.assertEqual(get_weekday_num("InvalidDay"), -1)



# test_db_helpers.py
# Path: appointment/tests/utils/test_db_helpers.py

import datetime
from unittest.mock import MagicMock, PropertyMock, patch

from django.apps import apps
from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import FieldDoesNotExist
from django.test import TestCase
from django.test.client import RequestFactory
from django.urls import reverse
from django.utils import timezone
from django_q.models import Schedule

from appointment.models import DayOff, PaymentInfo
from appointment.tests.base.base_test import BaseTest
from appointment.tests.mixins.base_mixin import ConfigMixin
from appointment.utils.db_helpers import (
    Config, WorkingHours, calculate_slots, calculate_staff_slots, can_appointment_be_rescheduled,
    cancel_existing_reminder, check_day_off_for_staff,
    create_and_save_appointment, create_new_user, create_payment_info_and_get_url, create_user_with_email,
    day_off_exists_for_date_range,
    exclude_booked_slots, exclude_pending_reschedules, generate_unique_username_from_email, get_absolute_url_,
    get_all_appointments,
    get_all_staff_members,
    get_appointment_buffer_time, get_appointment_by_id, get_appointment_finish_time, get_appointment_lead_time,
    get_appointment_slot_duration, get_appointments_for_date_and_time, get_config, get_day_off_by_id,
    get_non_working_days_for_staff, get_staff_member_appointment_list, get_staff_member_buffer_time,
    get_staff_member_by_user_id, get_staff_member_end_time, get_staff_member_from_user_id_or_logged_in,
    get_staff_member_slot_duration, get_staff_member_start_time, get_times_from_config, get_user_by_email,
    get_user_model, get_website_name, get_weekday_num_from_date, get_working_hours_by_id,
    get_working_hours_for_staff_and_day, is_working_day, parse_name, schedule_email_reminder,
    staff_change_allowed_on_reschedule, update_appointment_reminder, username_in_user_model, working_hours_exist
)


class TestCalculateSlots(TestCase):
    def setUp(self):
        self.start_time = datetime.datetime(2023, 10, 8, 8, 0)  # 8:00 AM
        self.end_time = datetime.datetime(2023, 10, 8, 12, 0)  # 12:00 PM
        self.slot_duration = datetime.timedelta(hours=1)  # 1 hour
        # Buffer time should've been datetime.datetime.now() but for the purpose of the tests, we'll use a fixed time.
        self.buffer_time = datetime.datetime(2023, 10, 8, 8, 0) + self.slot_duration

    def test_multiple_slots(self):
        """Buffer time goes 1 hour after the start time, it should only return three slots.
           Start time: 08:00 AM\n
           End time: 12:00 AM\n
           Buffer time: 09:00 AM\n
           Slot duration: 1 hour\n
        """

        expected = [
            datetime.datetime(2023, 10, 8, 9, 0),
            datetime.datetime(2023, 10, 8, 10, 0),
            datetime.datetime(2023, 10, 8, 11, 0)
        ]
        result = calculate_slots(self.start_time, self.end_time, self.buffer_time, self.slot_duration)
        self.assertEqual(result, expected)

    def test_buffer_after_end_time(self):
        """Buffer time goes beyond the end time, it should not then return any slots.
           Start time: 08:00 AM\n
           End time: 09:00 AM\n
           Buffer time: 10:00 AM\n
        """
        end_time = datetime.datetime(2023, 10, 8, 9, 0)
        buffer_time = datetime.datetime(2023, 10, 8, 10, 0)

        expected = []
        result = calculate_slots(self.start_time, end_time, buffer_time, self.slot_duration)
        self.assertEqual(result, expected)

    def test_one_slot_available(self):
        """Buffer time goes beyond the end time, it should not then return any slots.
           Start time: 08:00 AM\n
           End time: 09:00 AM\n
           Buffer time: 10:00 AM\n
        """
        end_time = datetime.datetime(2023, 10, 8, 9, 0)
        buffer_time = datetime.datetime(2023, 10, 8, 7, 30)
        slot_duration = datetime.timedelta(minutes=30)

        expected = [datetime.datetime(2023, 10, 8, 8, 0), datetime.datetime(2023, 10, 8, 8, 30)]
        result = calculate_slots(self.start_time, end_time, buffer_time, slot_duration)
        self.assertEqual(result, expected)


class TestCalculateStaffSlots(BaseTest):
    def setUp(self):
        super().setUp()
        self.slot_duration = datetime.timedelta(minutes=30)
        # Not working today but tomorrow
        self.date_not_working = datetime.date.today()
        self.working_date = datetime.date.today() + datetime.timedelta(days=1)
        weekday_num = get_weekday_num_from_date(self.working_date)
        WorkingHours.objects.create(
            staff_member=self.staff_member1,
            day_of_week=weekday_num,
            start_time=datetime.time(9, 0),
            end_time=datetime.time(17, 0)
        )
        self.staff_member1.appointment_buffer_time = 25.0
        self.buffer_time = datetime.timedelta(minutes=25)

    def test_calculate_slots_on_working_day_without_appointments(self):
        slots = calculate_staff_slots(self.working_date, self.staff_member1)
        # Slot duration is 30 minutes, so 8 working hours minus 25-minute buffer, divided by slot duration
        expected_number_of_slots = int((8 * 60 - 25) / 30)
        # 15 slots should be available instead of 16 because of the 25-minute buffer
        self.assertEqual(len(slots), expected_number_of_slots)

        # Asserting the first slot starts at 9:30 AM because of the 25-minute buffer
        self.assertEqual(slots[0].time(), datetime.time(9, 30))

        # Asserting the last slot starts before the end time minus slot duration (16:30)
        self.assertTrue((datetime.datetime.combine(self.working_date, slots[-1].time()) +
                         self.slot_duration).time() <= datetime.time(17, 0))

    def test_calculate_slots_on_non_working_day(self):
        """Test that no slots are returned on a day the staff member is not working."""
        slots = calculate_staff_slots(self.date_not_working, self.staff_member1)
        self.assertEqual(slots, [])


class TestCheckDayOffForStaff(BaseTest):
    def setUp(self):
        super().setUp()  # Call the parent class setup
        # Specific setups for this test class
        self.day_off1 = DayOff.objects.create(staff_member=self.staff_member1, start_date="2023-10-08",
                                              end_date="2023-10-10")
        self.day_off2 = DayOff.objects.create(staff_member=self.staff_member2, start_date="2023-10-05",
                                              end_date="2023-10-05")

    def test_staff_member_has_day_off(self):
        # Test for a date within the range of days off for staff_member1
        self.assertTrue(check_day_off_for_staff(self.staff_member1, "2023-10-09"))

    def test_staff_member_does_not_have_day_off(self):
        # Test for a date outside the range of days off for staff_member1
        self.assertFalse(check_day_off_for_staff(self.staff_member1, "2023-10-11"))

    def test_another_staff_member_day_off(self):
        # Test for a date within the range of days off for staff_member2
        self.assertTrue(check_day_off_for_staff(self.staff_member2, "2023-10-05"))

    def test_another_staff_member_no_day_off(self):
        # Test for a date outside the range of days off for staff_member2
        self.assertFalse(check_day_off_for_staff(self.staff_member2, "2023-10-06"))


class TestCreateAndSaveAppointment(BaseTest):

    def setUp(self):
        super().setUp()  # Call the parent class setup
        # Specific setups for this test class
        self.ar = self.create_appt_request_for_sm1()
        self.factory = RequestFactory()
        self.request = self.factory.get('/')

    def test_create_and_save_appointment(self):
        client_data = {
            'email': 'tester2@gmail.com',
            'name': 'Tester2',
        }
        appointment_data = {
            'phone': '123456789',
            'want_reminder': True,
            'address': '123 Main St',
            'additional_info': 'Additional Test Info'
        }

        appointment = create_and_save_appointment(self.ar, client_data, appointment_data, self.request)

        self.assertIsNotNone(appointment)
        self.assertEqual(appointment.client.email, client_data['email'])
        self.assertEqual(appointment.phone, appointment_data['phone'])
        self.assertEqual(appointment.want_reminder, appointment_data['want_reminder'])
        self.assertEqual(appointment.address, appointment_data['address'])
        self.assertEqual(appointment.additional_info, appointment_data['additional_info'])


def get_mock_reverse(url_name, **kwargs):
    """A mocked version of the reverse function."""
    if url_name == "app:view":
        return f'/mocked-url/{kwargs["kwargs"]["object_id"]}/{kwargs["kwargs"]["id_request"]}/'
    return reverse(url_name, **kwargs)


class ScheduleEmailReminderTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.factory = RequestFactory()
        self.request = self.factory.get('/')

    def test_schedule_email_reminder_cluster_running(self):
        appointment = self.create_appointment_for_user1()
        with patch('appointment.settings.check_q_cluster', return_value=True), \
                patch('appointment.utils.db_helpers.schedule') as mock_schedule:
            schedule_email_reminder(appointment, self.request)
            mock_schedule.assert_called_once()
            # Further assertions can be made here based on the arguments passed to schedule

    def test_schedule_email_reminder_cluster_not_running(self):
        appointment = self.create_appointment_for_user2()
        with patch('appointment.settings.check_q_cluster', return_value=False), \
                patch('appointment.utils.db_helpers.logger') as mock_logger:
            schedule_email_reminder(appointment, self.request)
            mock_logger.warning.assert_called_with(
                "Django-Q cluster is not running. Email reminder will not be scheduled.")


class UpdateAppointmentReminderTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.factory = RequestFactory()
        self.request = self.factory.get('/')
        self.appointment = self.create_appointment_for_user1()

    def test_update_appointment_reminder_date_time_changed(self):
        appointment = self.create_appointment_for_user1()
        new_date = timezone.now().date() + timezone.timedelta(days=10)
        new_start_time = timezone.now().time()

        with patch('appointment.utils.db_helpers.schedule_email_reminder') as mock_schedule_email_reminder, \
                patch('appointment.utils.db_helpers.cancel_existing_reminder') as mock_cancel_existing_reminder:
            update_appointment_reminder(appointment, new_date, new_start_time, self.request, True)
            mock_cancel_existing_reminder.assert_called_once_with(appointment.id_request)
            mock_schedule_email_reminder.assert_called_once()

    def test_update_appointment_reminder_no_change(self):
        appointment = self.create_appointment_for_user2()
        # Use existing date and time
        new_date = appointment.appointment_request.date
        new_start_time = appointment.appointment_request.start_time

        with patch('appointment.utils.db_helpers.schedule_email_reminder') as mock_schedule_email_reminder, \
                patch('appointment.utils.db_helpers.cancel_existing_reminder') as mock_cancel_existing_reminder:
            update_appointment_reminder(appointment, new_date, new_start_time, self.request, appointment.want_reminder)
            mock_cancel_existing_reminder.assert_not_called()
            mock_schedule_email_reminder.assert_not_called()

    @patch('appointment.utils.db_helpers.logger')  # Adjust the import path as necessary
    def test_reminder_not_scheduled_due_to_user_preference(self, mock_logger):
        # Scenario where user does not want a reminder
        want_reminder = False
        new_date = timezone.now().date() + datetime.timedelta(days=1)
        new_start_time = timezone.now().time()

        update_appointment_reminder(self.appointment, new_date, new_start_time, self.request, want_reminder)

        # Check that the logger.info was called with the expected message
        mock_logger.info.assert_called_once_with(
            f"Reminder for appointment {self.appointment.id} is not scheduled per user's preference or past datetime."
        )

    @patch('appointment.utils.db_helpers.logger')  # Adjust the import path as necessary
    def test_reminder_not_scheduled_due_to_past_datetime(self, mock_logger):
        # Scenario where the new datetime is in the past
        want_reminder = True
        new_date = timezone.now().date() - datetime.timedelta(days=1)  # Date in the past
        new_start_time = timezone.now().time()

        update_appointment_reminder(self.appointment, new_date, new_start_time, self.request, want_reminder)

        # Check that the logger.info was called with the expected message
        mock_logger.info.assert_called_once_with(
            f"Reminder for appointment {self.appointment.id} is not scheduled per user's preference or past datetime."
        )


# Helper method for modifying service rescheduling settings
def modify_service_rescheduling(service, **kwargs):
    for key, value in kwargs.items():
        setattr(service, key, value)
    service.save()


class CanAppointmentBeRescheduledTests(BaseTest, ConfigMixin):
    def setUp(self):
        super().setUp()
        self.appointment = self.create_appointment_for_user1()

    @patch('appointment.models.Service.reschedule_limit', new_callable=PropertyMock)
    @patch('appointment.models.Config.default_reschedule_limit', new=3)
    def test_can_appointment_be_rescheduled(self, mock_reschedule_limit):
        mock_reschedule_limit.return_value = 3
        self.assertTrue(can_appointment_be_rescheduled(self.appointment.appointment_request))

    def test_appointment_cannot_be_rescheduled_due_to_service_limit(self):
        modify_service_rescheduling(self.service1, allow_rescheduling=True, reschedule_limit=0)
        self.assertFalse(can_appointment_be_rescheduled(self.appointment.appointment_request))

    def test_rescheduling_allowed_exceeds_limit(self):
        modify_service_rescheduling(self.service1, allow_rescheduling=True, reschedule_limit=3)
        ar = self.create_appointment_request_with_histories(service=self.service1, count=4)
        self.assertFalse(can_appointment_be_rescheduled(ar))

    def test_rescheduling_with_default_limit(self):
        ar = self.create_appointment_request_with_histories(service=self.service1, count=2, use_default_limit=True)
        self.assertTrue(can_appointment_be_rescheduled(ar))
        self.create_appointment_reschedule_for_user1(appointment_request=ar)
        self.assertFalse(can_appointment_be_rescheduled(ar))

    # Helper method to create appointment request with rescheduled histories
    def create_appointment_request_with_histories(self, service, count, use_default_limit=False):
        ar = self.create_appointment_request_(service=service, staff_member=self.staff_member1)
        for _ in range(count):
            self.create_appointment_reschedule_for_user1(appointment_request=ar)
        return ar


class StaffChangeAllowedOnRescheduleTests(TestCase):
    @patch('appointment.models.Config.objects.first')
    def test_staff_change_allowed(self, mock_config_first):
        # Mock the Config object to return True for allow_staff_change_on_reschedule
        mock_config = MagicMock()
        mock_config.allow_staff_change_on_reschedule = True
        mock_config_first.return_value = mock_config

        # Call the function and assert that staff change is allowed
        self.assertTrue(staff_change_allowed_on_reschedule())

    @patch('appointment.models.Config.objects.first')
    def test_staff_change_not_allowed(self, mock_config_first):
        # Mock the Config object to return False for allow_staff_change_on_reschedule
        mock_config = MagicMock()
        mock_config.allow_staff_change_on_reschedule = False
        mock_config_first.return_value = mock_config

        # Call the function and assert that staff change is not allowed
        self.assertFalse(staff_change_allowed_on_reschedule())


class CreateUserWithEmailTests(TestCase):
    def setUp(self):
        self.User = get_user_model()
        self.User.USERNAME_FIELD = 'email'

    def test_create_user_with_full_data(self):
        """Test creating a user with complete client data."""
        client_data = {
            'email': 'test@example.com',
            'first_name': 'Test',
            'last_name': 'User',
            'username': 'test_user',
        }
        user = create_user_with_email(client_data)

        # Verify that the user was created with the correct attributes
        self.assertEqual(user.email, 'test@example.com')
        self.assertEqual(user.first_name, 'Test')
        self.assertEqual(user.last_name, 'User')
        self.assertTrue(user.is_active)
        self.assertFalse(user.is_staff)
        self.assertFalse(user.is_superuser)

    def test_create_user_with_partial_data(self):
        """Test creating a user with only an email provided."""
        client_data = {
            'email': 'partial@example.com',
            'username': 'partial_user',
            # First name and last name are omitted
        }
        user = create_user_with_email(client_data)

        # Verify that the user was created with default values for missing attributes
        self.assertEqual(user.email, 'partial@example.com')
        self.assertEqual(user.first_name, '')
        self.assertEqual(user.last_name, '')

    def test_create_user_with_duplicate_email(self):
        """Test attempting to create a user with a duplicate email."""
        client_data = {
            'email': 'duplicate@example.com',
            'first_name': 'Original',
            'last_name': 'User',
            'username': 'original_user',
        }
        # Create the first user
        create_user_with_email(client_data)

        # Attempt to create another user with the same email
        with self.assertRaises(Exception) as context:
            create_user_with_email(client_data)

        # Verify that the expected exception is raised (e.g., IntegrityError for duplicate key)
        self.assertTrue('duplicate key value violates unique constraint' in str(context.exception) or
                        'UNIQUE constraint failed' in str(context.exception))


class CancelExistingReminderTest(BaseTest):
    def test_cancel_existing_reminder(self):
        appointment = self.create_appointment_for_user1()
        Schedule.objects.create(func='appointment.tasks.send_email_reminder', name=f"reminder_{appointment.id_request}")

        self.assertEqual(Schedule.objects.count(), 1)
        cancel_existing_reminder(appointment.id_request)
        self.assertEqual(Schedule.objects.filter(name=f"reminder_{appointment.id_request}").count(), 0)


class TestCreatePaymentInfoAndGetUrl(BaseTest):

    def setUp(self):
        super().setUp()  # Call the parent class setup
        # Specific setups for this test class
        self.ar = self.create_appt_request_for_sm1()
        self.appointment = self.create_appointment_for_user2(appointment_request=self.ar)

    def test_create_payment_info_and_get_url_string(self):
        expected_url = "https://payment.com/1/1234567890"
        with patch('appointment.utils.db_helpers.APPOINTMENT_PAYMENT_URL', expected_url):
            payment_url = create_payment_info_and_get_url(self.appointment)
            self.assertEqual(payment_url, expected_url)

    def test_create_payment_info_and_get_url_application(self):
        expected_url = "app:view"

        with patch('appointment.utils.db_helpers.APPOINTMENT_PAYMENT_URL', expected_url):
            with patch('appointment.utils.db_helpers.reverse', side_effect=get_mock_reverse):
                self.assertEqual(PaymentInfo.objects.count(), 0)

                # Call the function to create PaymentInfo and get the URL
                payment_url = create_payment_info_and_get_url(self.appointment)

                # Now, there should be one PaymentInfo object
                self.assertEqual(PaymentInfo.objects.count(), 1)

                # Fetch the newly created PaymentInfo object
                payment_info = PaymentInfo.objects.first()

                # Construct the expected mocked URL
                expected_mocked_url = f'/mocked-url/{payment_info.id}/{payment_info.get_id_request()}/'

                # Assert that the appointment in the PaymentInfo object matches the appointment we provided
                self.assertEqual(payment_info.appointment, self.appointment)
                self.assertEqual(payment_url, expected_mocked_url)


class TestExcludeBookedSlots(BaseTest):

    def setUp(self):
        super().setUp()
        self.appointment = self.create_appointment_for_user1()

        # Sample slots for testing
        self.today = datetime.date.today()

        self.slots = [
            datetime.datetime.combine(self.today, datetime.time(8, 0)),
            datetime.datetime.combine(self.today, datetime.time(9, 0)),
            datetime.datetime.combine(self.today, datetime.time(10, 0)),
            datetime.datetime.combine(self.today, datetime.time(11, 0)),
            datetime.datetime.combine(self.today, datetime.time(12, 0))
        ]
        self.slot_duration = datetime.timedelta(hours=1)

    def test_no_appointments(self):
        result = exclude_booked_slots([], self.slots, self.slot_duration)
        self.assertEqual(result, self.slots)

    def test_appointment_not_intersecting_slots(self):
        self.appointment.appointment_request.start_time = datetime.time(13, 30)
        self.appointment.appointment_request.end_time = datetime.time(14, 30)
        self.appointment.save()

        result = exclude_booked_slots([self.appointment], self.slots, self.slot_duration)
        self.assertEqual(result, self.slots)

    def test_appointment_intersecting_single_slot(self):
        self.appointment.appointment_request.start_time = datetime.time(8, 0)
        self.appointment.appointment_request.end_time = datetime.time(9, 0)
        self.appointment.save()

        result = exclude_booked_slots([self.appointment], self.slots, self.slot_duration)
        expected = [
            datetime.datetime.combine(self.today, datetime.time(9, 0)),
            datetime.datetime.combine(self.today, datetime.time(10, 0)),
            datetime.datetime.combine(self.today, datetime.time(11, 0)),
            datetime.datetime.combine(self.today, datetime.time(12, 0))
        ]
        self.assertEqual(result, expected)

    def test_multiple_overlapping_appointments(self):
        ar2 = self.create_appt_request_for_sm2(start_time=datetime.time(10, 30),
                                               end_time=datetime.time(11, 30))
        appointment2 = self.create_appointment_for_user2(appointment_request=ar2)
        appointment2.save()
        result = exclude_booked_slots([self.appointment, appointment2], self.slots, self.slot_duration)
        expected = [
            datetime.datetime.combine(self.today, datetime.time(8, 0)),
            datetime.datetime.combine(self.today, datetime.time(12, 0))
        ]
        self.assertEqual(result, expected)


class TestDayOffExistsForDateRange(BaseTest):

    def setUp(self):
        super().setUp()
        self.user = self.create_user_(email="tester@gmail.com")
        self.service = self.create_service_()
        self.staff_member = self.create_staff_member_(user=self.user, service=self.service)
        self.day_off1 = DayOff.objects.create(staff_member=self.staff_member, start_date="2023-10-08",
                                              end_date="2023-10-10")
        self.day_off2 = DayOff.objects.create(staff_member=self.staff_member, start_date="2023-10-15",
                                              end_date="2023-10-17")

    def test_day_off_exists(self):
        # Check for a date range that intersects with day_off1
        self.assertTrue(day_off_exists_for_date_range(self.staff_member, "2023-10-09", "2023-10-11"))

    def test_day_off_does_not_exist(self):
        # Check for a date range that doesn't intersect with any day off
        self.assertFalse(day_off_exists_for_date_range(self.staff_member, "2023-10-11", "2023-10-14"))

    def test_day_off_exists_but_excluded(self):
        # Check for a date range that intersects with day_off1 but exclude day_off1 from the check using its ID
        self.assertFalse(
            day_off_exists_for_date_range(self.staff_member, "2023-10-09", "2023-10-11", days_off_id=self.day_off1.id))

    def test_day_off_exists_for_other_range(self):
        # Check for a date range that intersects with day_off2
        self.assertTrue(day_off_exists_for_date_range(self.staff_member, "2023-10-16", "2023-10-18"))


class TestGetAllAppointments(BaseTest):

    def setUp(self):
        super().setUp()
        self.appointment1 = self.create_appointment_for_user1()
        self.appointment2 = self.create_appointment_for_user2()

    def test_get_all_appointments(self):
        appointments = get_all_appointments()
        self.assertEqual(len(appointments), 2)
        self.assertIn(self.appointment1, appointments)
        self.assertIn(self.appointment2, appointments)


class TestGetAllStaffMembers(BaseTest):

    def setUp(self):
        super().setUp()

    def test_get_all_staff_members(self):
        staff_members = get_all_staff_members()
        self.assertEqual(len(staff_members), 2)
        self.assertIn(self.staff_member1, staff_members)
        self.assertIn(self.staff_member2, staff_members)


class TestGetAppointmentByID(BaseTest):

    def setUp(self):
        super().setUp()
        self.appointment = self.create_appointment_for_user1()

    def test_existing_appointment(self):
        """Test fetching an existing appointment."""
        fetched_appointment = get_appointment_by_id(self.appointment.id)
        self.assertEqual(fetched_appointment, self.appointment)

    def test_non_existing_appointment(self):
        """Test attempting to fetch a non-existing appointment."""
        non_existent_id = 9999  # Assume this ID doesn't exist in the test database
        fetched_appointment = get_appointment_by_id(non_existent_id)
        self.assertIsNone(fetched_appointment)


@patch('appointment.utils.db_helpers.APPOINTMENT_BUFFER_TIME', 60)
@patch('appointment.utils.db_helpers.APPOINTMENT_LEAD_TIME', '07:00:00')
@patch('appointment.utils.db_helpers.APPOINTMENT_FINISH_TIME', '15:00:00')
@patch('appointment.utils.db_helpers.APPOINTMENT_SLOT_DURATION', 30)
@patch('appointment.utils.db_helpers.APPOINTMENT_WEBSITE_NAME', "django-appointment-website")
class TestGetAppointmentConfigTimes(TestCase):
    def test_no_config_object(self):
        """Test when there's no Config object in the database."""
        self.assertIsNone(Config.objects.first())  # Ensure no Config object exists
        self.assertEqual(get_appointment_buffer_time(), 60)
        self.assertEqual(get_appointment_lead_time(), '07:00:00')
        self.assertEqual(get_appointment_finish_time(), '15:00:00')
        self.assertEqual(get_appointment_slot_duration(), 30)
        self.assertEqual(get_website_name(), "django-appointment-website")

    def test_config_object_no_time_set(self):
        """Test with a Config object without 'slot_duration'; 'buffer', 'lead' and 'finish' time set."""
        Config.objects.create()
        self.assertEqual(get_appointment_buffer_time(), 60)
        self.assertEqual(get_appointment_lead_time(), '07:00:00')
        self.assertEqual(get_appointment_finish_time(), '15:00:00')
        self.assertEqual(get_appointment_slot_duration(), 30)
        self.assertEqual(get_website_name(), "django-appointment-website")

    def test_config_object_with_finish_time(self):
        """Test with a Config object with 'slot_duration'; 'buffer', 'lead' and 'finish' time set."""
        Config.objects.create(finish_time='17:00:00', lead_time='09:00:00',
                              appointment_buffer_time=60, slot_duration=30, website_name="config")
        self.assertEqual(get_appointment_buffer_time(), 60)
        self.assertEqual(get_appointment_lead_time().strftime('%H:%M:%S'), '09:00:00')
        self.assertEqual(get_appointment_finish_time().strftime('%H:%M:%S'), '17:00:00')
        self.assertEqual(get_appointment_slot_duration(), 30)
        self.assertEqual(get_website_name(), "config")

    def test_config_not_set_but_constants_patched(self):
        """Test with no Config object and patched constants."""
        self.assertEqual(get_appointment_buffer_time(), 60)
        self.assertEqual(get_appointment_lead_time(), '07:00:00')
        self.assertEqual(get_appointment_finish_time(), '15:00:00')
        self.assertEqual(get_appointment_slot_duration(), 30)
        self.assertEqual(get_website_name(), "django-appointment-website")


class TestGetAppointmentsForDateAndTime(BaseTest):

    def setUp(self):
        super().setUp()

        # Setting up some appointment requests and appointments for testing
        self.date_sample = datetime.datetime.today()

        # Creating overlapping appointments for staff_member1
        ar1 = self.create_appt_request_for_sm1(start_time=datetime.time(9, 0), end_time=datetime.time(10, 0))
        self.appointment1 = self.create_appointment_(user=self.client1, appointment_request=ar1)

        ar2 = self.create_appt_request_for_sm1(start_time=datetime.time(10, 30), end_time=datetime.time(11, 30))
        self.appointment2 = self.create_appointment_(user=self.client2, appointment_request=ar2)

        # Creating a non-overlapping appointment for staff_member1
        ar3 = self.create_appt_request_for_sm1(start_time=datetime.time(13, 0), end_time=datetime.time(14, 0))
        self.appointment3 = self.create_appointment_(user=self.client1, appointment_request=ar3)

    def test_get_appointments_overlapping_time_range(self):
        """Test retrieving appointments overlapping with a specific time range."""
        appointments = get_appointments_for_date_and_time(self.date_sample, datetime.time(10, 0), datetime.time(12, 0),
                                                          self.staff_member1)
        self.assertEqual(appointments.count(), 2)
        self.assertIn(self.appointment1, appointments)
        self.assertIn(self.appointment2, appointments)

    def test_get_appointments_outside_time_range(self):
        """Test retrieving appointments outside a specific time range."""
        appointments = get_appointments_for_date_and_time(self.date_sample, datetime.time(7, 0), datetime.time(8, 30),
                                                          self.staff_member1)
        self.assertEqual(appointments.count(), 0)

    def test_get_appointments_for_different_date(self):
        """Test retrieving appointments for a different date."""
        appointments = get_appointments_for_date_and_time(datetime.date(2023, 10, 11), datetime.time(9, 0),
                                                          datetime.time(12, 0), self.staff_member1)
        self.assertEqual(appointments.count(), 0)

    def test_get_appointments_for_different_staff_member(self):
        """Test retrieving appointments for a different staff member."""
        appointments = get_appointments_for_date_and_time(self.date_sample, datetime.time(9, 0), datetime.time(12, 0),
                                                          self.staff_member2)
        self.assertEqual(appointments.count(), 0)


class TestGetConfig(TestCase):

    def setUp(self):
        # Clear the cache at the start of each test to ensure a clean state
        cache.clear()

    def test_no_config_in_cache_or_db(self):
        """Test when there's no Config in cache or the database."""
        config = get_config()
        self.assertIsNone(config)

    def test_config_in_db_not_in_cache(self):
        """Test when there's a Config object in the database but not in the cache."""
        db_config = Config.objects.create(finish_time='17:00:00')
        config = get_config()
        self.assertEqual(config, db_config)

    def test_config_in_cache(self):
        """Test when there's a Config object in the cache."""
        db_config = Config.objects.create(finish_time='17:00:00')
        cache.set('config', db_config)
        # Use 'patch' to ensure the Config model isn't hit during the test
        with patch('appointment.models.Config.objects.first') as mock_first:
            config = get_config()
            self.assertEqual(config, db_config)
            mock_first.assert_not_called()  # Ensure the database wasn't hit


class TestGetDayOffById(BaseTest):  # Assuming you have a BaseTest class with some initial setups
    def setUp(self):
        super().setUp()  # Call the parent class setup

        # Assuming you have already set up some StaffMember objects in the BaseTest
        self.day_off = DayOff.objects.create(staff_member=self.staff_member1, start_date="2023-10-08",
                                             end_date="2023-10-10")

    def test_retrieve_existing_day_off(self):
        """Test retrieving an existing DayOff object."""
        retrieved_day_off = get_day_off_by_id(self.day_off.id)
        self.assertEqual(retrieved_day_off, self.day_off)

    def test_nonexistent_day_off_id(self):
        """Test trying to retrieve a DayOff object using a non-existent ID."""
        nonexistent_id = self.day_off.id + 1  # Just to ensure a non-existent ID, you can use any logic that suits you
        retrieved_day_off = get_day_off_by_id(nonexistent_id)
        self.assertIsNone(retrieved_day_off)


class TestGetNonWorkingDaysForStaff(BaseTest):
    def setUp(self):
        super().setUp()  # Call the parent class setup

        self.staff_member_with_working_days = self.staff_member1
        WorkingHours.objects.create(staff_member=self.staff_member_with_working_days, day_of_week=0,
                                    start_time=datetime.time(8, 0), end_time=datetime.time(12, 0))
        WorkingHours.objects.create(staff_member=self.staff_member_with_working_days, day_of_week=2,
                                    start_time=datetime.time(8, 0), end_time=datetime.time(12, 0))

        self.staff_member_without_working_days = self.staff_member2

    def test_retrieve_non_working_days_for_staff_with_some_working_days(self):
        """Test retrieving non-working days for a StaffMember with some working days set."""
        non_working_days = get_non_working_days_for_staff(self.staff_member_with_working_days.id)
        expected_days = [1, 3, 4, 5, 6]
        self.assertListEqual(non_working_days, expected_days)

    def test_retrieve_non_working_days_for_staff_without_working_days(self):
        """Test retrieving non-working days for a StaffMember with no working days set."""
        non_working_days = get_non_working_days_for_staff(self.staff_member_without_working_days.id)
        expected_days = [0, 1, 2, 3, 4, 5, 6]
        self.assertListEqual(non_working_days, expected_days)

    def test_nonexistent_staff_member_id(self):
        """Test trying to retrieve non-working days using a non-existent StaffMember ID."""
        nonexistent_id = self.staff_member_with_working_days.id + 100  # Just to ensure a non-existent ID
        non_working_days = get_non_working_days_for_staff(nonexistent_id)
        self.assertListEqual(non_working_days, [])


class TestGetStaffMemberAppointmentList(BaseTest):
    def setUp(self):
        super().setUp()

        # Creating appointments for each staff member.
        self.appointment1_for_user1 = self.create_appointment_for_user1()
        self.appointment2_for_user2 = self.create_appointment_for_user2()

    def test_retrieve_appointments_for_specific_staff_member(self):
        """Test retrieving appointments for a specific StaffMember."""
        # Testing for staff_member1
        appointments_for_staff_member1 = get_staff_member_appointment_list(self.staff_member1)
        self.assertIn(self.appointment1_for_user1, appointments_for_staff_member1)
        self.assertNotIn(self.appointment2_for_user2, appointments_for_staff_member1)

        # Testing for staff_member2
        appointments_for_staff_member2 = get_staff_member_appointment_list(self.staff_member2)
        self.assertIn(self.appointment2_for_user2, appointments_for_staff_member2)
        self.assertNotIn(self.appointment1_for_user1, appointments_for_staff_member2)

    def test_retrieve_appointments_for_staff_member_with_no_appointments(self):
        """Test retrieving appointments for a StaffMember with no appointments."""
        # Creating a new staff member with no appointments
        staff_member_with_no_appointments = self.create_staff_member_(user=self.client1, service=self.service1)
        appointments = get_staff_member_appointment_list(staff_member_with_no_appointments)
        self.assertListEqual(list(appointments), [])


class TestGetWeekdayNumFromDate(TestCase):
    def test_get_weekday_num_from_date(self):
        """Test getting the weekday number from a date."""
        sample_dates = {
            datetime.date(2023, 10, 9): 1,  # Monday
            datetime.date(2023, 10, 10): 2,  # Tuesday
            datetime.date(2023, 10, 11): 3,  # Wednesday
            datetime.date(2023, 10, 12): 4,  # Thursday
            datetime.date(2023, 10, 13): 5,  # Friday
            datetime.date(2023, 10, 14): 6,  # Saturday
            datetime.date(2023, 10, 15): 0,  # Sunday
        }

        for date, expected_weekday_num in sample_dates.items():
            self.assertEqual(get_weekday_num_from_date(date), expected_weekday_num)


@patch('appointment.utils.db_helpers.APPOINTMENT_BUFFER_TIME', 60)
@patch('appointment.utils.db_helpers.APPOINTMENT_SLOT_DURATION', 30)
class TestStaffMemberTimeFunctions(BaseTest):
    """Test suite for staff member time functions."""

    def setUp(self):
        super().setUp()

        # Set staff member-specific settings
        self.staff_member1.slot_duration = 15
        self.staff_member1.lead_time = datetime.time(8, 30)
        self.staff_member1.finish_time = datetime.time(18, 0)
        self.staff_member1.appointment_buffer_time = 45
        self.staff_member1.save()

        # Setting WorkingHours for staff_member1 for Monday
        self.wh = WorkingHours.objects.create(
            staff_member=self.staff_member1,
            day_of_week=1,
            start_time=datetime.time(9, 0),
            end_time=datetime.time(17, 0)
        )

    def test_staff_member_buffer_time_with_global_setting(self):
        """Test buffer time when staff member-specific setting is None."""
        self.staff_member1.appointment_buffer_time = None
        self.staff_member1.save()
        buffer_time = get_staff_member_buffer_time(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(buffer_time, 60)  # Global setting

    def test_staff_member_buffer_time_with_staff_member_setting(self):
        """Test buffer time using staff member-specific setting."""
        buffer_time = get_staff_member_buffer_time(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(buffer_time, 45)  # Staff member specific setting

    def test_staff_member_slot_duration_with_global_setting(self):
        """Test slot duration when staff member-specific setting is None."""
        self.staff_member1.slot_duration = None
        self.staff_member1.save()
        slot_duration = get_staff_member_slot_duration(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(slot_duration, 30)  # Global setting

    def test_staff_member_slot_duration_with_staff_member_setting(self):
        """Test slot duration using staff member-specific setting."""
        slot_duration = get_staff_member_slot_duration(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(slot_duration, 15)  # Staff member specific setting

    def test_staff_member_start_time(self):
        """Test start time based on WorkingHours."""
        start_time = get_staff_member_start_time(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(start_time, datetime.time(9, 0))

    def test_staff_member_end_time(self):
        """Test end time based on WorkingHours."""
        end_time = get_staff_member_end_time(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(end_time, datetime.time(17, 0))

    def test_staff_member_start_time_with_lead_time(self):
        """Test start time when both lead_time and WorkingHours are available."""
        start_time = get_staff_member_start_time(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(start_time, self.wh.start_time)  # lead_time should prevail

    def test_staff_member_end_time_with_finish_time(self):
        """Test end time when both finish_time and WorkingHours are available."""
        end_time = get_staff_member_end_time(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(end_time, self.wh.end_time)  # finish_time should prevail

    def test_staff_member_buffer_time_with_working_hours_conflict(self):
        """Test buffer time when it conflicts with WorkingHours."""
        self.staff_member1.appointment_buffer_time = 120  # Set a buffer time greater than WorkingHours start time
        self.staff_member1.save()
        buffer_time = get_staff_member_buffer_time(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(buffer_time, 120)  # Should still use staff member-specific setting even if it causes conflict


class TestDBHelpers(BaseTest):
    def setUp(self):
        super().setUp()

    def test_get_staff_member_by_user_id(self):
        """Test retrieving a StaffMember object using a user id works as expected."""
        staff = get_staff_member_by_user_id(self.user1.id)
        self.assertIsNotNone(staff)
        self.assertEqual(staff, self.staff_member1)

        # Test for a non-existent user id
        staff = get_staff_member_by_user_id(99999)
        self.assertIsNone(staff)

    def test_get_staff_member_from_user_id_or_logged_in(self):
        """Test retrieving a StaffMember object using a user id or the logged in user works as expected."""
        staff = get_staff_member_from_user_id_or_logged_in(self.user1)
        self.assertIsNotNone(staff)
        self.assertEqual(staff, self.staff_member1)

        staff = get_staff_member_from_user_id_or_logged_in(self.user1, self.user2.id)
        self.assertIsNotNone(staff)
        self.assertEqual(staff, self.staff_member2)

        # Test for a non-existent user id
        staff = get_staff_member_from_user_id_or_logged_in(self.user1, 99999)
        self.assertIsNone(staff)

    def test_get_user_model(self):
        """Test retrieving the user model works as expected."""
        user_model = get_user_model()
        user_model_in_settings = apps.get_model(settings.AUTH_USER_MODEL)
        self.assertEqual(user_model, user_model_in_settings)

    def test_get_user_by_email(self):
        """Retrieve a user by email."""
        user = get_user_by_email("tester1@gmail.com")
        self.assertIsNotNone(user)
        self.assertEqual(user, self.user1)

        # Test for a non-existent email
        user = get_user_by_email("nonexistent@gmail.com")
        self.assertIsNone(user)


class TestWorkingHoursFunctions(BaseTest):
    def setUp(self):
        super().setUp()
        self.working_hours = WorkingHours.objects.create(
            staff_member=self.staff_member1,
            day_of_week=1,  # Monday
            start_time=datetime.time(9, 0),
            end_time=datetime.time(17, 0)
        )

    def test_get_working_hours_by_id(self):
        """Test retrieving a WorkingHours object by ID."""
        working_hours = get_working_hours_by_id(self.working_hours.id)
        self.assertEqual(working_hours, self.working_hours)

        # Non-existent ID
        working_hours = get_working_hours_by_id(99999)
        self.assertIsNone(working_hours)

    def test_get_working_hours_for_staff_and_day(self):
        """Test retrieving WorkingHours for a specific staff member and day."""
        # With set WorkingHours
        result = get_working_hours_for_staff_and_day(self.staff_member1, 1)
        self.assertEqual(result['start_time'], datetime.time(9, 0))
        self.assertEqual(result['end_time'], datetime.time(17, 0))

        # Without set WorkingHours but with staff member's default times
        self.staff_member1.lead_time = datetime.time(8, 0)
        self.staff_member1.finish_time = datetime.time(18, 0)
        self.staff_member1.save()
        result = get_working_hours_for_staff_and_day(self.staff_member1, 2)  # Tuesday
        self.assertEqual(result['start_time'], datetime.time(8, 0))
        self.assertEqual(result['end_time'], datetime.time(18, 0))

    def test_is_working_day(self):
        """is_working_day() should return True if there are WorkingHours for the staff member and day,
           False otherwise."""
        self.assertTrue(is_working_day(self.staff_member1, 1))  # Monday
        self.assertFalse(is_working_day(self.staff_member1, 2))  # Tuesday

    def test_working_hours_exist(self):
        """working_hours_exist() should return True if there are WorkingHours for the staff member and day,"""
        self.assertTrue(working_hours_exist(1, self.staff_member1))  # Monday
        self.assertFalse(working_hours_exist(2, self.staff_member1))  # Tuesday


@patch('appointment.utils.db_helpers.APPOINTMENT_LEAD_TIME', (7, 0))
@patch('appointment.utils.db_helpers.APPOINTMENT_FINISH_TIME', (15, 0))
@patch('appointment.utils.db_helpers.APPOINTMENT_SLOT_DURATION', 30)
@patch('appointment.utils.db_helpers.APPOINTMENT_BUFFER_TIME', 60)
class TestGetTimesFromConfig(TestCase):
    def setUp(self):
        self.sample_date = datetime.date(2023, 10, 9)
        cache.clear()

    def test_times_from_config_object(self):
        """Test retrieving times from a Config object."""
        # Create a Config object with custom values
        Config.objects.create(
            lead_time=datetime.time(9, 0),
            finish_time=datetime.time(17, 0),
            slot_duration=45,
            appointment_buffer_time=90
        )

        start_time, end_time, slot_duration, buff_time = get_times_from_config(self.sample_date)

        # Assert times from 'Config' object
        self.assertEqual(start_time, datetime.datetime(2023, 10, 9, 9, 0))
        self.assertEqual(end_time, datetime.datetime(2023, 10, 9, 17, 0))
        self.assertEqual(slot_duration, datetime.timedelta(minutes=45))
        self.assertEqual(buff_time, datetime.timedelta(minutes=90))

    def test_times_from_default_settings(self):
        """Test retrieving times from default settings."""
        # Ensure no Config object exists
        Config.objects.all().delete()

        start_time, end_time, slot_duration, buff_time = get_times_from_config(self.sample_date)

        # Assert times from default settings
        self.assertEqual(start_time, datetime.datetime(2023, 10, 9, 7, 0))
        self.assertEqual(end_time, datetime.datetime(2023, 10, 9, 15, 0))
        self.assertEqual(slot_duration, datetime.timedelta(minutes=30))
        self.assertEqual(buff_time, datetime.timedelta(minutes=60))


class CreateNewUserTest(TestCase):
    def test_create_new_user_unique_username(self):
        """Test creating a new user with a unique username."""
        client_data = {'name': 'John Doe', 'email': 'john.doe@example.com'}
        user = create_new_user(client_data)
        self.assertEqual(user.username, 'john.doe')
        self.assertEqual(user.first_name, 'John')
        self.assertEqual(user.email, 'john.doe@example.com')

    def test_create_new_user_duplicate_username(self):
        """Test creating a new user with a duplicate username."""
        client_data1 = {'name': 'John Doe', 'email': 'john.doe@example.com'}
        user1 = create_new_user(client_data1)
        self.assertEqual(user1.username, 'john.doe')

        client_data2 = {'name': 'Jane Doe', 'email': 'john.doe@example.com'}
        user2 = create_new_user(client_data2)
        self.assertEqual(user2.username, 'john.doe01')  # Suffix added

        client_data3 = {'name': 'James Doe', 'email': 'john.doe@example.com'}
        user3 = create_new_user(client_data3)
        self.assertEqual(user3.username, 'john.doe02')  # Next suffix

    def test_generate_unique_username(self):
        """Test if generate_unique_username_from_email function generates unique usernames."""
        email = 'john.doe@example.com'
        username = generate_unique_username_from_email(email)
        self.assertEqual(username, 'john.doe')

        # Assuming we have a user with the same username
        CLIENT_MODEL = get_user_model()
        CLIENT_MODEL.objects.create_user(username='john.doe', email=email)
        new_username = generate_unique_username_from_email(email)
        self.assertEqual(new_username, 'john.doe01')

    def test_parse_name(self):
        """Test if parse_name function splits names correctly."""
        name = "John Doe"
        first_name, last_name = parse_name(name)
        self.assertEqual(first_name, 'John')
        self.assertEqual(last_name, 'Doe')

    def test_create_new_user_check_password(self):
        """Test creating a new user with a password."""
        client_data = {'name': 'John Doe', 'email': 'john.doe@example.com'}
        user = create_new_user(client_data)
        # Check that no password has been set
        self.assertFalse(user.has_usable_password())


class UsernameInUserModelTests(TestCase):

    @patch('django.contrib.auth.models.User._meta.get_field')
    def test_username_field_exists(self, mock_get_field):
        """
        Test that `username_in_user_model` returns True when the 'username' field exists.
        """
        mock_get_field.return_value = True  # Mocking as if 'username' field exists
        self.assertTrue(username_in_user_model())

    @patch('django.contrib.auth.models.User._meta.get_field')
    def test_username_field_does_not_exist(self, mock_get_field):
        """
        Test that `username_in_user_model` returns False when the 'username' field does not exist.
        """
        mock_get_field.side_effect = FieldDoesNotExist  # Simulating 'username' field does not exist
        self.assertFalse(username_in_user_model())


class ExcludePendingReschedulesTests(BaseTest):

    def setUp(self):
        super().setUp()
        self.date = timezone.now().date() + datetime.timedelta(minutes=5)
        self.start_time = (timezone.now() - datetime.timedelta(minutes=4)).time()
        self.end_time = (timezone.now() + datetime.timedelta(minutes=1)).time()

        self.slots = [
            datetime.datetime.combine(self.date, self.start_time),
            datetime.datetime.combine(self.date, self.end_time)
        ]

    def test_exclude_no_pending_reschedules(self):
        """Slots should remain unchanged if there are no pending rescheduling."""
        filtered_slots = exclude_pending_reschedules(self.slots, self.staff_member1, self.date)
        self.assertEqual(len(filtered_slots), len(self.slots))

    def test_exclude_with_pending_reschedules_outside_last_5_minutes(self):
        """Slots should remain unchanged if pending reschedules are outside the last 5 minutes."""
        appointment_request = self.create_appointment_request_(self.service1, self.staff_member1)
        self.create_reschedule_history_(
            appointment_request,
            date_=self.date,
            start_time=(timezone.now() - datetime.timedelta(minutes=10)).time(),
            end_time=(timezone.now() - datetime.timedelta(minutes=5)).time(),
            staff_member=self.staff_member1,
            reason_for_rescheduling="Scheduling conflict"
        )
        filtered_slots = exclude_pending_reschedules(self.slots, self.staff_member1, self.date)
        self.assertEqual(len(filtered_slots), len(self.slots))

    def test_exclude_with_pending_reschedules_within_last_5_minutes(self):
        """Slots overlapping with pending rescheduling within the last 5 minutes should be excluded."""
        appointment_request = self.create_appointment_request_(self.service1, self.staff_member1)
        reschedule_start_time = (timezone.now() - datetime.timedelta(minutes=4)).time()
        reschedule_end_time = (timezone.now() + datetime.timedelta(minutes=1)).time()
        self.create_reschedule_history_(
            appointment_request,
            date_=self.date,
            start_time=reschedule_start_time,
            end_time=reschedule_end_time,
            staff_member=self.staff_member1,
            reason_for_rescheduling="Client request"
        )
        filtered_slots = exclude_pending_reschedules(self.slots, self.staff_member1, self.date)
        self.assertEqual(len(filtered_slots), len(self.slots) - 1)  # Assuming only one slot overlaps

    def test_exclude_with_non_pending_reschedules_within_last_5_minutes(self):
        """Slots should remain unchanged if reschedules within the last 5 minutes are not pending."""
        appointment_request = self.create_appointment_request_(self.service1, self.staff_member1)
        reschedule_start_time = (timezone.now() - datetime.timedelta(minutes=4)).time()
        reschedule_end_time = (timezone.now() + datetime.timedelta(minutes=1)).time()
        reschedule = self.create_reschedule_history_(
            appointment_request,
            date_=self.date,
            start_time=reschedule_start_time,
            end_time=reschedule_end_time,
            staff_member=self.staff_member1,
            reason_for_rescheduling="Urgent issue"
        )
        reschedule.reschedule_status = 'confirmed'
        reschedule.save()
        filtered_slots = exclude_pending_reschedules(self.slots, self.staff_member1, self.date)
        self.assertEqual(len(filtered_slots), len(self.slots))


class GetAbsoluteUrlTests(TestCase):

    def setUp(self):
        # Create a RequestFactory instance
        self.factory = RequestFactory()

    def test_get_absolute_url_with_request(self):
        # Create a request object using RequestFactory
        request = self.factory.get('/some-path/')
        relative_url = '/test-url/'
        expected_url = request.build_absolute_uri(relative_url)

        # Call the function with the request object
        result_url = get_absolute_url_(relative_url, request)

        # Assert the result matches the expected URL
        self.assertEqual(result_url, expected_url)



# db_helpers.py
# Path: appointment/utils/db_helpers.py

"""
Author: Adams Pierre David
Since: 2.0.0
"""

import datetime
from typing import Optional
from urllib.parse import urlparse

from django.apps import apps
from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import FieldDoesNotExist
from django.urls import reverse
from django.utils import timezone
from django_q.models import Schedule
from django_q.tasks import schedule

from appointment.logger_config import logger
from appointment.settings import (
    APPOINTMENT_BUFFER_TIME, APPOINTMENT_FINISH_TIME, APPOINTMENT_LEAD_TIME, APPOINTMENT_PAYMENT_URL,
    APPOINTMENT_SLOT_DURATION, APPOINTMENT_WEBSITE_NAME
)
from appointment.utils.date_time import combine_date_and_time, get_weekday_num

Appointment = apps.get_model('appointment', 'Appointment')
AppointmentRequest = apps.get_model('appointment', 'AppointmentRequest')
WorkingHours = apps.get_model('appointment', 'WorkingHours')
DayOff = apps.get_model('appointment', 'DayOff')
PaymentInfo = apps.get_model('appointment', 'PaymentInfo')
StaffMember = apps.get_model('appointment', 'StaffMember')
Config = apps.get_model('appointment', 'Config')
Service = apps.get_model('appointment', 'Service')
EmailVerificationCode = apps.get_model('appointment', 'EmailVerificationCode')
AppointmentRescheduleHistory = apps.get_model('appointment', 'AppointmentRescheduleHistory')


def calculate_slots(start_time, end_time, buffer_time, slot_duration):
    """Calculate the available slots between the given start and end times using the given buffer time and slot duration

    :param start_time: The start time.
    :param end_time: The end time.
    :param buffer_time: The buffer time.
    :param slot_duration: The duration of each slot.
    :return: A list of available slots.
    """
    slots = []
    buffer_time = buffer_time.replace(tzinfo=None)
    while start_time + slot_duration <= end_time:
        if start_time >= buffer_time:
            slots.append(start_time)
        start_time += slot_duration
    return slots


def calculate_staff_slots(date, staff_member):
    """Calculate the available slots for the given staff member on the given date.

    :param date: The date to calculate the slots for.
    :param staff_member: The staff member to calculate the slots for.
    :return: A list of available slots.
    """
    # Convert the times to datetime objects
    weekday_num = get_weekday_num_from_date(date)
    if not is_working_day(staff_member, weekday_num):
        return []
    staff_member_start_time = get_staff_member_start_time(staff_member, date)
    start_time = datetime.datetime.combine(date, staff_member_start_time)
    end_time = datetime.datetime.combine(date, get_staff_member_end_time(staff_member, date))

    # Convert the buffer duration in minutes to a timedelta object
    buffer_duration_minutes = get_staff_member_buffer_time(staff_member, date)
    buffer_duration = datetime.timedelta(minutes=buffer_duration_minutes)
    buffer_time_init = datetime.datetime.combine(date, staff_member_start_time)
    buffer_time = buffer_time_init + buffer_duration

    # Convert slot duration to a timedelta object
    slot_duration_minutes = get_staff_member_slot_duration(staff_member, date)
    slot_duration = datetime.timedelta(minutes=slot_duration_minutes)

    return calculate_slots(start_time, end_time, buffer_time, slot_duration)


def check_day_off_for_staff(staff_member, date) -> bool:
    """Check if the given staff member is off on the given date.
    :param staff_member: The staff member to check.
    :param date: The date to check.
    """
    return DayOff.objects.filter(staff_member=staff_member, start_date__lte=date, end_date__gte=date).exists()


def create_and_save_appointment(ar, client_data: dict, appointment_data: dict, request):
    """Create and save a new appointment based on the provided appointment request and client data.

    :param ar: The appointment request associated with the new appointment.
    :param client_data: The data of the client making the appointment.
    :param appointment_data: Additional data for the appointment, including phone number, address, etc.
    :param request: The request object.
    :return: The newly created appointment.
    """
    user = get_user_by_email(client_data['email'])
    appointment = Appointment.objects.create(
        client=user, appointment_request=ar,
        **appointment_data
    )
    appointment.save()
    logger.info(f"New appointment created: {appointment.to_dict()}")
    schedule_email_reminder(appointment, request)
    return appointment


def schedule_email_reminder(appointment, request, appointment_datetime=None):
    """Schedule an email reminder for the given appointment."""
    # Check if the Django-Q cluster is running
    from appointment.settings import check_q_cluster
    if not check_q_cluster():
        logger.warning("Django-Q cluster is not running. Email reminder will not be scheduled.")
        return

    # Calculate reminder datetime if not provided
    if appointment_datetime is None:
        appointment_datetime = combine_date_and_time(appointment.appointment_request.date,
                                                     appointment.appointment_request.start_time)
    if timezone.is_naive(appointment_datetime):
        appointment_datetime = timezone.make_aware(appointment_datetime)

    reminder_datetime = appointment_datetime - datetime.timedelta(days=1)

    ar_id_request = appointment.appointment_request.get_id_request()
    relative_reschedule_url = reverse('appointment:prepare_reschedule_appointment', args=[ar_id_request])
    reschedule_link = get_absolute_url_(relative_reschedule_url, request)

    logger.info(f"Scheduling email reminder for appointment {appointment.id} at {reminder_datetime}")

    # Schedule the email reminder task with Django-Q
    schedule('appointment.tasks.send_email_reminder',
             to_email=appointment.client.email,
             name=f"reminder_{appointment.id_request}",
             first_name=appointment.client.first_name,
             reschedule_link=reschedule_link,
             appointment_id=appointment.id,
             schedule_type=Schedule.ONCE,  # Use Schedule.ONCE for a one-time task
             next_run=reminder_datetime)


def update_appointment_reminder(appointment, new_date, new_start_time, request, want_reminder=None):
    """
    Updates or cancels the appointment reminder based on changes to the start time or date,
    and the user's preference for receiving a reminder.
    """
    # Convert new date and time strings to datetime objects for comparison
    new_datetime = combine_date_and_time(new_date, new_start_time)

    existing_datetime = combine_date_and_time(appointment.appointment_request.date,
                                              appointment.appointment_request.start_time)

    # Ensure new_datetime is timezone-aware
    if timezone.is_naive(new_datetime):
        new_datetime = timezone.make_aware(new_datetime)

    # Ensure existing_datetime is timezone-aware
    if timezone.is_naive(existing_datetime):
        existing_datetime = timezone.make_aware(existing_datetime)

    # Determine if there's been a change in the datetime or the reminder preference
    want_reminder = want_reminder if want_reminder is not None else appointment.want_reminder
    datetime_changed = new_datetime != existing_datetime
    reminder_preference_changed = appointment.want_reminder != want_reminder

    if datetime_changed or reminder_preference_changed:
        # Cancel any existing reminder
        cancel_existing_reminder(appointment.id_request)

        # If a reminder is still desired and the appointment is in the future, schedule a new one
        if want_reminder and new_datetime > timezone.now():
            schedule_email_reminder(appointment, request, new_datetime)
        else:
            logger.info(
                f"Reminder for appointment {appointment.id} is not scheduled per user's preference or past datetime.")

    # Update the appointment's reminder preference
    appointment.want_reminder = want_reminder
    appointment.save()


def cancel_existing_reminder(appointment_id_request):
    """
    Cancels any existing reminder for the appointment.
    """
    task_name = f"reminder_{appointment_id_request}"
    Schedule.objects.filter(name=task_name).delete()


def can_appointment_be_rescheduled(appointment_request):
    # Datetime 5 minutes ago from now
    five_minutes_ago = timezone.now() - timezone.timedelta(minutes=5)

    # Filter reschedule histories to those created within the last 5 minutes
    recent_reschedule_count = appointment_request.reschedule_histories.filter(created_at__gte=five_minutes_ago).count()
    service = appointment_request.service
    config = Config.get_instance()

    # Determine which rescheduled limit to use based on service settings
    if service.allow_rescheduling:
        # If rescheduling is allowed
        logger.info(f"Rescheduling is allowed for service {service.name} -> "
                    f"Reschedule count: {recent_reschedule_count}, Reschedule limit: {service.reschedule_limit}")
        return recent_reschedule_count < service.reschedule_limit
    else:
        # Rescheduling is allowed but no specific limit set; use system default
        logger.info(f"Rescheduling is allowed but no specific limit set for service {service.name} -> "
                    f"Reschedule count: {recent_reschedule_count}, Reschedule limit: {config.default_reschedule_limit}")
        return recent_reschedule_count < config.default_reschedule_limit


def staff_change_allowed_on_reschedule():
    return Config.objects.first().allow_staff_change_on_reschedule


def generate_unique_username_from_email(email: str) -> str:
    username_base = email.split('@')[0]
    username = username_base
    suffix = 1
    CLIENT_MODEL = get_user_model()

    while CLIENT_MODEL.objects.filter(username=username).exists():
        suffix_str = f"{suffix:02}"
        username = f"{username_base}{suffix_str}"
        suffix += 1
    return username


def parse_name(name: str):
    return name.split(' ', 1)


def create_user_with_email(client_data: dict):
    CLIENT_MODEL = get_user_model()
    # Valid fields
    valid_fields = ['email', 'first_name', 'last_name', 'username']

    # Filter client_data to include only valid fields
    user_data = {field: client_data.get(field, '') for field in valid_fields}

    user = CLIENT_MODEL.objects.create_user(**user_data)
    return user


def create_user_with_username(client_data: dict):
    CLIENT_MODEL = get_user_model()
    if 'username' not in client_data:
        username = generate_unique_username_from_email(client_data['email'])
    else:
        username = client_data['username']
    user_data = {
        'username': username,
        'email': client_data['email'],
        'first_name': client_data.get('first_name', ''),
        'last_name': client_data.get('last_name', '')
    }
    user = CLIENT_MODEL.objects.create_user(**user_data)
    return user


def create_new_user(client_data: dict):
    CLIENT_MODEL = get_user_model()

    # Check if client_data has first_name and last_name
    if 'first_name' not in client_data or 'last_name' not in client_data:
        # Assuming 'name' contains a single space between first and last name.
        client_data['first_name'], client_data['last_name'] = parse_name(client_data['name'])

    try:
        # Check if the 'username' field exists in the User model
        CLIENT_MODEL._meta.get_field('username')
        user = create_user_with_username(client_data)
    except FieldDoesNotExist:
        user = create_user_with_email(client_data)
    return user


def username_in_user_model():
    CLIENT_MODEL = get_user_model()
    try:
        # Check if the 'username' field exists in the User model
        CLIENT_MODEL._meta.get_field('username')
        return True
    except FieldDoesNotExist:
        return False


def create_payment_info_and_get_url(appointment):
    """
    Create a new payment information entry for the appointment and return the payment URL.

    :param appointment: The appointment to create the payment information for.
    :return: The payment URL for the appointment.
    """
    # Create a new PaymentInfo entry for the appointment
    payment_info = PaymentInfo(appointment=appointment)
    payment_info.save()

    # Check if APPOINTMENT_PAYMENT_URL is a Django reverse URL (e.g. "app:view_name") or an external link
    if (":" in APPOINTMENT_PAYMENT_URL and "/" not in APPOINTMENT_PAYMENT_URL) and not bool(
            urlparse(APPOINTMENT_PAYMENT_URL).netloc):
        # It's a Django reverse URL; generate the URL
        payment_url = reverse(
            APPOINTMENT_PAYMENT_URL,
            kwargs={'object_id': payment_info.id, 'id_request': payment_info.get_id_request()}
        )
    else:
        # It's an external link; return as is or append necessary data
        payment_url = APPOINTMENT_PAYMENT_URL

    return payment_url


def exclude_booked_slots(appointments, slots, slot_duration=None):
    """Exclude the booked slots from the given list of slots.

    :param appointments: The appointments to exclude.
    :param slots: The slots to exclude the appointments from.
    :param slot_duration: The duration of each slot.
    :return: The slots with the booked slots excluded.
    """
    available_slots = []
    for slot in slots:
        slot_end = slot + slot_duration
        is_available = True
        for appointment in appointments:
            appointment_start_time = appointment.get_start_time()
            appointment_end_time = appointment.get_end_time()
            if appointment_start_time < slot_end and slot < appointment_end_time:
                is_available = False
                break
        if is_available:
            available_slots.append(slot)
    return available_slots


def exclude_pending_reschedules(slots, staff_member, date):
    """
    Exclude the slots that are pending reschedule for the given staff member and date.
    """

    # Calculate the time window for "last 5 minutes"
    ten_minutes_ago = timezone.now() - datetime.timedelta(minutes=5)
    pending_reschedules = AppointmentRescheduleHistory.objects.filter(
        appointment_request__staff_member=staff_member,
        date=date,
        reschedule_status='pending',
        created_at__gte=ten_minutes_ago
    )

    # Filter out slots that overlap with any pending rescheduling
    filtered_slots = slots[:]
    for reschedule in pending_reschedules:
        reschedule_start_time = datetime.datetime.combine(date, reschedule.start_time)
        reschedule_end_time = datetime.datetime.combine(date, reschedule.end_time)

        filtered_slots = [slot for slot in filtered_slots if
                          not (reschedule_start_time <= slot < reschedule_end_time)]

    return filtered_slots


def day_off_exists_for_date_range(staff_member, start_date, end_date, days_off_id=None) -> bool:
    """Check if a day off exists for the given staff member and date range.

    :param staff_member: The staff member to check.
    :param start_date: The start date of the date range.
    :param end_date: The end date of the date range.
    :param days_off_id: The ID of the day off to exclude from the check.
    :return: True if a day off exists for the given staff member and date range; otherwise, False.
    """
    days_off = DayOff.objects.filter(staff_member=staff_member, start_date__lte=end_date, end_date__gte=start_date)
    if days_off_id:
        days_off = days_off.exclude(id=days_off_id)
    return days_off.exists()


def get_all_appointments() -> list:
    """Get all appointments from the database.

    :return: QuerySet, all appointments
    """
    return Appointment.objects.all()


def get_all_staff_members() -> list:
    """Get all staff members from the database.

    :return: QuerySet, all staff members
    """
    return StaffMember.objects.all()


def get_appointment_buffer_time():
    """Get the appointment buffer time from the settings file.

    :return: The appointment buffer time
    """
    from appointment.models import Config

    config = Config.objects.first()

    if config and config.appointment_buffer_time:
        return config.appointment_buffer_time
    return APPOINTMENT_BUFFER_TIME


def get_appointment_by_id(appointment_id):
    """Get an appointment by its ID.

    :param appointment_id: The appointment's ID
    :return: The appointment with the specified ID or None if no appointment with the specified ID exists
    """
    try:
        return Appointment.objects.get(id=appointment_id)
    except Appointment.DoesNotExist:
        return None


def get_appointment_finish_time():
    """Get the appointment finish time from the settings file.

    :return: The appointment's finish time
    """
    from appointment.models import Config

    config = Config.objects.first()

    if config and config.finish_time:
        return config.finish_time
    return APPOINTMENT_FINISH_TIME


def get_appointment_lead_time():
    """Get the appointment lead time from the settings file.

    :return: The appointment's lead time
    """
    from appointment.models import Config

    config = Config.objects.first()

    if config and config.lead_time:
        return config.lead_time
    return APPOINTMENT_LEAD_TIME


def get_appointment_slot_duration():
    """Get the appointment slot duration from the settings file.

    :return: The appointment slot duration
    """
    from appointment.models import Config

    config = Config.objects.first()

    if config and config.slot_duration:
        return config.slot_duration
    return APPOINTMENT_SLOT_DURATION


def get_appointments_for_date_and_time(date, start_time, end_time, staff_member):
    """Returns all appointments that overlap with the specified date and time range.

    :param date: The date to filter appointments on.
    :param start_time: The starting time to filter appointments on.
    :param end_time: The ending time to filter appointments on.
    :param staff_member: The staff member to filter appointments on.

    :return: QuerySet, all appointments that overlap with the specified date and time range
    """
    return Appointment.objects.filter(
        appointment_request__date=date,
        appointment_request__start_time__lte=end_time,
        appointment_request__end_time__gte=start_time,
        appointment_request__staff_member=staff_member
    )


def get_config():
    """Returns the configuration object from the database or the cache."""
    config = cache.get('config')
    if not config:
        config = Config.objects.first()
        # Cache the configuration for 1 hour (3600 seconds)
        cache.set('config', config, 3600)
    return config


def get_day_off_by_id(day_off_id):
    """Get a day off by its ID.

    :param day_off_id: The day offs ID
    :return: DayOff, the day off with the specified ID or None if no day off with the specified ID exists.
    """
    try:
        return DayOff.objects.get(pk=day_off_id)
    except DayOff.DoesNotExist:
        return None


def get_non_working_days_for_staff(staff_member_id):
    """Return the non-working days for the given staff member or an empty list if the staff member does not exist."""
    all_days = set(range(7))  # Represents all days (0-6)
    try:
        staff_member = StaffMember.objects.get(id=staff_member_id)
        working_days = set(WorkingHours.objects.filter(staff_member=staff_member).values_list('day_of_week', flat=True))

        # Subtracting working_days from all_days to get non-working days
        non_working_days = list(all_days - working_days)
        return non_working_days
    except StaffMember.DoesNotExist:
        return []


def get_staff_member_appointment_list(staff_member: StaffMember) -> list:
    """Get a list of appointments for the given staff member."""
    return Appointment.objects.filter(appointment_request__staff_member=staff_member)


def get_weekday_num_from_date(date: datetime.date = None) -> int:
    """Get the number of the weekday from the given date."""
    if date is None:
        date = datetime.date.today()
    return get_weekday_num(date.strftime("%A"))


def get_staff_member_buffer_time(staff_member: StaffMember, date: datetime.date) -> float:
    """Return the buffer time for the given staff member on the given date."""
    _, _, _, buff_time = get_times_from_config(date)
    buffer_minutes = buff_time.total_seconds() / 60
    return staff_member.appointment_buffer_time or buffer_minutes


def get_staff_member_by_user_id(user_id):
    """Return a staff member by their user ID."""
    try:
        return StaffMember.objects.get(user_id=user_id)
    except StaffMember.DoesNotExist:
        return None


def get_staff_member_end_time(staff_member: StaffMember, date: datetime.date) -> Optional[datetime.time]:
    """Return the end time for the given staff member on the given date."""
    weekday_num = get_weekday_num_from_date(date)
    working_hours = get_working_hours_for_staff_and_day(staff_member, weekday_num)
    return working_hours['end_time']


def get_staff_member_from_user_id_or_logged_in(user, user_id=None):
    """Fetch StaffMember based on the user_id or the logged-in user."""
    staff_member = None
    try:
        if user_id:
            staff_member = StaffMember.objects.get(user_id=user_id)
        else:
            staff_member = user.staffmember
    except StaffMember.DoesNotExist:
        pass
    return staff_member


def get_staff_member_slot_duration(staff_member: StaffMember, date: datetime.date) -> int:
    """Return the slot duration for the given staff member on the given date."""
    _, _, slot_duration, _ = get_times_from_config(date)
    slot_minutes = slot_duration.total_seconds() / 60
    return staff_member.slot_duration or slot_minutes


def get_staff_member_start_time(staff_member: StaffMember, date: datetime.date) -> Optional[datetime.time]:
    """Return the start time for the given staff member on the given date."""
    weekday_num = get_weekday_num_from_date(date)
    working_hours = get_working_hours_for_staff_and_day(staff_member, weekday_num)
    return working_hours['start_time']


def get_times_from_config(date):
    """Get the start time, end time, slot duration, and buffer time from the configuration or the settings file.

    :param date: The date to get the times for.
    :return: The start time, end time, slot duration, and buffer time.
    """
    config = get_config()
    if config:
        start_time = datetime.datetime.combine(date, datetime.time(hour=config.lead_time.hour,
                                                                   minute=config.lead_time.minute))
        end_time = datetime.datetime.combine(date, datetime.time(hour=config.finish_time.hour,
                                                                 minute=config.finish_time.minute))
        slot_duration = datetime.timedelta(minutes=config.slot_duration)
        buff_time = datetime.timedelta(minutes=config.appointment_buffer_time)
    else:
        start_hour, start_minute = APPOINTMENT_LEAD_TIME
        start_time = datetime.datetime.combine(date, datetime.time(hour=start_hour, minute=start_minute))
        finish_hour, finish_minute = APPOINTMENT_FINISH_TIME
        end_time = datetime.datetime.combine(date, datetime.time(hour=finish_hour, minute=finish_minute))
        slot_duration = datetime.timedelta(minutes=APPOINTMENT_SLOT_DURATION)
        buff_time = datetime.timedelta(minutes=APPOINTMENT_BUFFER_TIME)
    return start_time, end_time, slot_duration, buff_time


def get_user_model():
    """Get the client models from the settings file.

    :return: The user model
    """
    return apps.get_model(settings.AUTH_USER_MODEL)


def get_user_by_email(email: str):
    """Get a user by their email address.

    :param email: The email address of the user.
    :return: The user with the specified email address, if found; otherwise, None.
    """
    CLIENT_MODEL = get_user_model()
    return CLIENT_MODEL.objects.filter(email=email).first()


def get_website_name() -> str:
    """Get the website name from the configuration file.

    :return: The website name
    """
    from appointment.models import Config

    config = Config.objects.first()

    if config and config.website_name != "":
        return config.website_name
    return APPOINTMENT_WEBSITE_NAME


def get_working_hours_by_id(working_hours_id):
    """Get a working hours by its ID.

    :param working_hours_id: The working hours' ID
    :return: WorkingHours, the working hours with the specified ID
    """
    try:
        return WorkingHours.objects.get(pk=working_hours_id)
    except WorkingHours.DoesNotExist:
        return None


def get_working_hours_for_staff_and_day(staff_member, day_of_week):
    """Get the working hours for the given staff member and day of the week.

    :param staff_member: The staff member to get the working hours for.
    :param day_of_week: The day of the week to get the working hours for.
    :return: The working hours for the given staff member and day of the week.
    """
    working_hours = WorkingHours.objects.filter(staff_member=staff_member, day_of_week=day_of_week).first()
    start_time = staff_member.get_lead_time()
    end_time = staff_member.get_finish_time()
    if not working_hours and not (start_time and end_time):
        return None
    # If no specific working hours are set for that day, use the default start and end times from StaffMember
    if not working_hours:
        return {
            'staff_member': staff_member,
            'day_of_week': day_of_week,
            'start_time': staff_member.get_lead_time(),
            'end_time': staff_member.get_finish_time()
        }

    # If a WorkingHours instance is found, convert it to a dictionary for consistent return type
    return {
        'staff_member': working_hours.staff_member,
        'day_of_week': working_hours.day_of_week,
        'start_time': working_hours.start_time,
        'end_time': working_hours.end_time
    }


def is_working_day(staff_member: StaffMember, day: int) -> bool:
    """Check if the given day is a working day for the staff member."""
    working_days = list(WorkingHours.objects.filter(staff_member=staff_member).values_list('day_of_week', flat=True))
    return day in working_days


def working_hours_exist(day_of_week, staff_member):
    """Check if working hours exist for the given day of the week and staff member."""
    return WorkingHours.objects.filter(day_of_week=day_of_week, staff_member=staff_member).exists()


def get_absolute_url_(relative_url, request):
    return request.build_absolute_uri(relative_url)
